from ..architectures.embeddings.similarity import SimilarityEmbedding
from ..callbacks import SaveAugmentedSimilarityBatch
from ..losses import VICRegLoss
from .base import AmplfiModel


class SimilarityModel(AmplfiModel):
    """
    A LightningModule for training similarity embeddings

    Args:
        arch:
            A neural network architecture that maps waveforms
            to lower dimensional embedded space
    """

    def __init__(
        self,
        *args,
        arch: SimilarityEmbedding,
        similarity_loss: VICRegLoss,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        # TODO: parmeterize cov, std, repr weights
        self.model = arch
        self.similarity_loss = similarity_loss

    def forward(
        self,
        ref,
        aug,
    ):
        ref = self.model(ref)
        aug = self.model(aug)
        loss, (inv_loss, var_loss, cov_loss) = self.similarity_loss(ref, aug)
        return loss, (inv_loss, var_loss, cov_loss)

    def validation_step(self, batch, _):
        [ref, aug], asds, _ = batch
        loss, (inv_loss, var_loss, cov_loss) = self((ref, asds), (aug, asds))
        self.log(
            "valid_loss", loss, on_epoch=True, prog_bar=True, sync_dist=True
        )

        self.log("valid_var_loss", var_loss, on_step=False, on_epoch=True)
        self.log("valid_cov_loss", cov_loss, on_step=False, on_epoch=True)
        self.log("valid_inv_loss", inv_loss, on_step=False, on_epoch=True)

        return loss

    def training_step(self, batch, _):
        # unpack batch - can ignore parameters
        [ref, aug], asds, _ = batch

        # pass reference and augmented data contexts
        # through embedding and calculate similarity loss
        loss, (inv_loss, var_loss, cov_loss) = self((ref, asds), (aug, asds))

        self.log(
            "train_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            sync_dist=True,
        )

        self.log("train_var_loss", var_loss, on_step=True, on_epoch=True)
        self.log("train_cov_loss", cov_loss, on_step=True, on_epoch=True)
        self.log("train_inv_loss", inv_loss, on_step=True, on_epoch=True)

        return loss

    def configure_callbacks(self):
        callbacks = super().configure_callbacks()
        callbacks.append(SaveAugmentedSimilarityBatch())
        return callbacks



import logging
import os
import sys
from typing import Dict, List, Optional, Sequence

import h5py
import lightning.pytorch as pl
import torch
from ml4gw.dataloading import Hdf5TimeSeriesDataset, InMemoryDataset
from ml4gw.transforms import ChannelWiseScaler, Whiten

from ...augmentations import PsdEstimator, WaveformProjector
from ..utils import fs as fs_utils
from ..utils.utils import ZippedDataset
from ..waveforms.sampler import WaveformSampler
import numpy as np
from pathlib import Path
import random

Tensor = torch.Tensor
Distribution = torch.distributions.Distribution


class AmplfiDataset(pl.LightningDataModule):
    """
    Base LightningDataModule for loading data to train AMPLFI models.

    Subclasses must define the `inject` method
    which encodes how background strain,
    cross/plus polarizations and parameters
    are processed before being passed to a model

    Args:
        data_dir:
            Path to directory containing training and testing data
        inference_params:
            List of parameters to perform inference on. Can be a subset
            of the parameters that fully describes the waveforms
        highpass:
            Highpass frequency in Hz
        sample_rate:
            Rate data is sampled in Hz
        kernel_length:
            Length of the kernel seen by model in seconds
        fduration:
            The length of the whitening filter's impulse
            response, in seconds. `fduration / 2` seconds
            worth of data will be cropped from the edges
            of the whitened timeseries.
        psd_length:
            Length of data used to calculate psd in seconds
        batches_per_epoch:
            Number of batches for each training epoch.
        batch_size:
            Number of samples in each batch
        ifos:
            List of interferometers
        waveform_sampler:
            `WaveformSampler` object that produces waveforms and parameters
            for training, validation and testing.
            See `train.data.waveforms.sampler`
            for methods this object should define.
        fftlength:
            Length of the fft used to calculate the psd.
            Defaults to `kernel_length`
        min_valid_duration:
            Minimum number of seconds of validation background data
        num_files_per_batch:
            Number of strain hdf5 files to use to construct
            each batch. Can lead to dataloading performance increases.
        max_num_workers:
            Maximum number of workers to assign to each
            training dataloader.

    """

    def __init__(
        self,
        data_dir: str,
        inference_params: list[str],
        highpass: float,
        sample_rate: float,
        kernel_length: float,
        fduration: float,
        psd_length: float,
        batches_per_epoch: int,
        batch_size: int,
        ifos: List[str],
        waveform_sampler: WaveformSampler,
        fftlength: Optional[int] = None,
        min_valid_duration: float = 10000,
        num_files_per_batch: Optional[int] = None,
        max_num_workers: int = 6,
        verbose: bool = False,
    ):
        super().__init__()
        self.save_hyperparameters(ignore=["waveform_sampler"])
        self.init_logging(verbose)
        self.waveform_sampler = waveform_sampler
        self.max_num_workers = max_num_workers

        # generate our local node data directory
        # if our specified data source is remote
        self.data_dir = fs_utils.get_data_dir(self.hparams.data_dir)

    def init_logging(self, verbose: bool):
        log_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        logging.basicConfig(
            format=log_format,
            level=logging.DEBUG if verbose else logging.INFO,
            stream=sys.stdout,
        )

    def prepare_data(self):
        """
        Download s3 data if it doesn't exist.
        """
        logger = logging.getLogger(self.__class__.__name__)
        bucket, _ = fs_utils.split_data_dir(self.hparams.data_dir)
        if bucket is None:
            return
        logger.info(
            "Downloading data from S3 bucket {} to {}".format(
                bucket, self.data_dir
            )
        )
        fs_utils.download_training_data(bucket, self.data_dir)

    # ================================================ #
    # Distribution utilities
    # ================================================ #
    def get_world_size_and_rank(self) -> tuple[int, int]:
        """
        Name says it all, but generalizes to the case
        where we aren't running distributed training.
        """
        if not torch.distributed.is_initialized():
            return 1, 0
        else:
            world_size = torch.distributed.get_world_size()
            rank = torch.distributed.get_rank()
            return world_size, rank

    def get_logger(self, world_size, rank):
        logger_name = self.__class__.__name__
        if world_size > 1:
            logger_name += f":{rank}"
        return logging.getLogger(logger_name)

    @property
    def device(self):
        """Return the device of the associated lightning module"""
        return self.trainer.lightning_module.device

    # ================================================ #
    # Helper utilities for preprocessing
    # ================================================ #

    def transform(self, parameters: Dict[str, Tensor]):
        """
        Make transforms to parameters before scaling
        and performing training/inference.
        For example, taking logarithm of hrss
        """
        return self.waveform_sampler.parameter_transformer(parameters)

    def scale(self, parameters, reverse: bool = False):
        """
        Apply standard scaling to transformed parameters
        """
        parameters = parameters.transpose(1, 0)
        scaled = self.scaler(parameters, reverse=reverse)
        scaled = scaled.transpose(1, 0)
        return scaled

    # ================================================ #
    # Re-parameterizing some attributes
    # ================================================ #

    @property
    def frequencies(self):
        """
        Frequencies corresponding to an fft
        of a timeseries of length `kernel_length`
        """
        size = int(self.hparams.kernel_length * self.hparams.sample_rate)
        return torch.fft.rfftfreq(size, d=1 / self.hparams.sample_rate)

    @property
    def sample_length(self):
        return (
            self.hparams.kernel_length
            + self.hparams.fduration
            + self.hparams.psd_length
        )

    @property
    def num_ifos(self):
        return len(self.hparams.ifos)

    @property
    def num_params(self):
        return len(self.hparams.inference_params)

    @property
    def num_workers(self):
        local_world_size = len(self.trainer.device_ids)
        return min(
            self.max_num_workers, int(os.cpu_count() / local_world_size)
        )

    @property
    def val_batch_size(self):
        """Use larger batch sizes when we don't need gradients."""
        return int(1 * self.hparams.batch_size)

    @property
    def train_val_fnames(self):
        """List of background files used for both training and validation"""
        background_dir = self.data_dir / "train" / "background"
        fnames = list(background_dir.glob("*.hdf5"))
        return fnames

    @property
    def test_fnames(self):
        """List of background files used for testing a trained model"""
        test_dir = self.data_dir / "test" / "background"
        fnames = list(test_dir.glob("*.hdf5"))
        return fnames

    def train_val_split(self) -> Sequence[str]:
        """
        Split background files into training and validation sets
        based on the requested duration of the validation set
        """
        fnames = sorted(self.train_val_fnames)
        durations = [int(fname.stem.split("-")[-1]) for fname in fnames]
        valid_fnames = []
        valid_duration = 0
        while valid_duration < self.hparams.min_valid_duration:
            fname, duration = fnames.pop(-1), durations.pop(-1)
            valid_duration += duration
            valid_fnames.append(str(fname))

        train_fnames = fnames
        return train_fnames, valid_fnames

    # ================================================ #
    # Utilities for initial data loading and preparation
    # ================================================ #

    def transforms_to_device(self):
        """
        Move all `torch.nn.Modules` to the local device
        """
        for item in self.__dict__.values():
            if isinstance(item, torch.nn.Module):
                item.to(self.device)

    def build_transforms(self, stage):
        """
        Build torch.nn.Modules that will be used for on-device
        augmentation and preprocessing. Transfer these modules
        to the appropiate device
        """
        self._logger.info("Building torch Modules and transferring to device")
        window_length = self.hparams.kernel_length + self.hparams.fduration
        fftlength = self.hparams.fftlength or window_length
        self.psd_estimator = PsdEstimator(
            window_length,
            self.hparams.sample_rate,
            fftlength,
            fast=self.hparams.highpass is not None,
            average="median",
        )

        self.whitener = Whiten(
            self.hparams.fduration,
            self.hparams.sample_rate,
            self.hparams.highpass,
        )

        # build standard scaler object and fit to parameters;
        # waveform_sampler subclasses will decide how to generate
        # parameters to fit the scaler.
        # Only fit the scaler during training.
        # During testing, use pre-fit parameters
        if stage in ["predict", "test"]:
            self._logger.info("Using pre-fit standard scaler")
            self.scaler = self.trainer.model.scaler
        else:
            self._logger.info("Fitting standard scaler to parameters")
            scaler = ChannelWiseScaler(self.num_params)
            self.scaler = self.waveform_sampler.fit_scaler(scaler)

        self.projector = WaveformProjector(
            self.hparams.ifos, self.hparams.sample_rate
        )

    def setup(self, stage: str) -> None:
        world_size, rank = self.get_world_size_and_rank()
        self._logger = self.get_logger(world_size, rank)
        self.train_fnames, self.val_fnames = self.train_val_split()

        self._logger.info(f"Setting up data for stage {stage}")

        # infer sample rate directly from background data
        with h5py.File(self.train_fnames[0], "r") as f:
            sample_rate = 1 / f[self.hparams.ifos[0]].attrs["dx"]
            assert sample_rate == self.hparams.sample_rate

        self._logger.info(f"Inferred sample rate of {sample_rate} Hz")

        # load validation/testing waveforms, parameters, and background
        # and build the fixed background dataset while data and augmentations
        # modules are all still on CPU.
        # get_val_waveforms should be implemented by waveform_sampler object
        if stage in ["fit", "validate"]:
            self.val_background = self.load_background(self.val_fnames)
            self._logger.info(
                f"Loaded background files {self.val_fnames} for validation"
            )

            cross, plus, parameters = self.waveform_sampler.get_val_waveforms(
                rank, world_size
            )
            self._logger.info(f"Loaded {len(cross)} waveforms for validation")
            params = []
            for k in self.hparams.inference_params:
                if k in parameters.keys():
                    params.append(torch.Tensor(parameters[k]))

            self.val_waveforms = torch.stack([cross, plus], dim=0)
            self.val_parameters = torch.column_stack(params)

        elif stage == "test":
            self._logger.info(
                f"Loaded background files {self.test_fnames} for testing"
            )
            (
                cross,
                plus,
                parameters,
            ) = self.waveform_sampler.get_test_waveforms()

            self._logger.info(f"Loaded {len(cross)} waveforms for testing")

            params = []
            for k in self.hparams.inference_params:
                if k in parameters.keys():
                    params.append(torch.Tensor(parameters[k]))

            self.test_waveforms = torch.stack([cross, plus], dim=0)
            self.test_parameters = torch.column_stack(params)

        # once we've generated validation/testing waveforms on cpu,
        # build data augmentation modules
        # and transfer them to appropiate device
        self.build_transforms(stage)
        self.transforms_to_device()

    def load_background(self, fnames: Sequence[str]):
        background = []
        for fname in fnames:
            data = []
            with h5py.File(fname, "r") as f:
                for ifo in self.hparams.ifos:
                    back = f[ifo][:]
                    data.append(torch.tensor(back, dtype=torch.float32))
            data = torch.stack(data)
            background.append(data)
        return background

    def on_after_batch_transfer(self, batch, _):
        """
        This is a Lightning `hook` that gets called after
        data returned by a dataloader gets put on the local device,
        but before it gets passed to model for inference.

        Use this to do on-device augmentation/preprocessing
        """
        if self.trainer.training:
            [batch] = batch
            cross, plus, parameters = self.waveform_sampler.sample(batch)
            strain, asds, parameters = self.inject(
                batch, cross, plus, parameters
            )

        elif self.trainer.validating or self.trainer.sanity_checking:
            [cross, plus, parameters], [background] = batch

            background = background[: len(cross)]
            keys = [
                k
                for k in self.hparams.inference_params
                if k not in ["dec", "psi", "phi"]
            ]
            parameters = {k: parameters[:, i] for i, k in enumerate(keys)}
            strain, asds, parameters = self.inject(
                background, cross, plus, parameters
            )

        elif self.trainer.testing:
            [cross, plus, parameters], [background] = batch
            keys = [
                k
                for k in self.hparams.inference_params
                if k not in ["dec", "psi", "phi"]
            ]
            parameters = {k: parameters[:, i] for i, k in enumerate(keys)}
            strain, asds, parameters = self.inject(
                background, cross, plus, parameters
            )

        return strain, asds, parameters

    # ================================================ #
    # Dataloaders used by lightning
    # ================================================ #

    def train_dataloader(self):
        # if we only have one training file
        # load it into memory and use InMemoryDataset
        if len(self.train_fnames) == 1:
            train_background = self.load_background(self.train_fnames)[0]
            dataset = InMemoryDataset(
                train_background,
                kernel_size=int(self.hparams.sample_rate * self.sample_length),
                batch_size=self.hparams.batch_size,
                coincident=False,
                batches_per_epoch=self.hparams.batches_per_epoch,
                shuffle=True,
            )
        else:
            dataset = Hdf5TimeSeriesDataset(
                self.train_fnames,
                channels=self.hparams.ifos,
                kernel_size=int(self.hparams.sample_rate * self.sample_length),
                batch_size=self.hparams.batch_size,
                batches_per_epoch=self.hparams.batches_per_epoch,
                coincident=False,
                num_files_per_batch=self.hparams.num_files_per_batch,
            )

        self._logger.info(
            f"Using a {dataset.__class__.__name__} class for training"
        )
        pin_memory = isinstance(
            self.trainer.accelerator, pl.accelerators.CUDAAccelerator
        )
        dataloader = torch.utils.data.DataLoader(
            dataset, num_workers=self.num_workers, pin_memory=pin_memory
        )
        return dataloader

    def val_dataloader(self):
        # TODO: allow for multiple validation segment files

        # offset the start of the validation background data
        # by the device id to add more diversity in the validation set
        _, rank = self.get_world_size_and_rank()

        # build waveform dataloader
        cross, plus = self.val_waveforms
        waveform_dataset = torch.utils.data.TensorDataset(
            cross, plus, self.val_parameters
        )
        waveform_dataloader = torch.utils.data.DataLoader(
            waveform_dataset,
            batch_size=self.val_batch_size,
            shuffle=False,
            pin_memory=False,
        )

        # build background dataloader
        val_background = self.val_background[0][:, rank:]

        background_dataset = InMemoryDataset(
            val_background,
            kernel_size=int(self.hparams.sample_rate * self.sample_length),
            batch_size=self.val_batch_size,
            batches_per_epoch=len(waveform_dataloader),
            coincident=False,
            shuffle=False,
        )

        background_dataloader = torch.utils.data.DataLoader(
            background_dataset, pin_memory=False
        )
        return ZippedDataset(
            waveform_dataloader,
            background_dataloader,
        )

    def test_dataloader(self):
        # build waveform dataloader
        cross, plus = self.test_waveforms
        waveform_dataset = torch.utils.data.TensorDataset(
            cross, plus, self.test_parameters
        )
        waveform_dataloader = torch.utils.data.DataLoader(
            waveform_dataset,
            batch_size=1,
            shuffle=False,
            pin_memory=False,
            num_workers=10,
        )

        if len(self.test_fnames) == 1:
            test_background = self.load_background(self.test_fnames)[0]
            background_dataset = InMemoryDataset(
                test_background,
                kernel_size=int(self.hparams.sample_rate * self.sample_length),
                batch_size=self.hparams.batch_size,
                coincident=False,
                batches_per_epoch=self.hparams.batches_per_epoch,
                shuffle=True,
            )
        else:
            background_dataset = Hdf5TimeSeriesDataset(
                self.test_fnames,
                channels=self.hparams.ifos,
                kernel_size=int(self.hparams.sample_rate * self.sample_length),
                batch_size=1,
                batches_per_epoch=len(waveform_dataloader),
                coincident=False,
            )

        background_dataloader = torch.utils.data.DataLoader(
            background_dataset, pin_memory=False, num_workers=10
        )
        return ZippedDataset(
            waveform_dataloader,
            background_dataloader,
        )

    def inject(self, *args, **kwargs):
        """
        Subclasses should implement this method
        for different training use cases,
        like training a similarity embedding
        or training a normalizing flow. This is called
        after the data is transferred to the local device
        """
        raise NotImplementedError

    def background_from_gpstimes(
        self, gpstimes: np.ndarray, fnames: List[Path]
    ) -> torch.Tensor:
        """
        Construct a Tensor of background segments corresponding
        to requested `gpstimes` where `gpstimes` specifies
        the desired location of the coalescence time
        """

        # load in background segments corresponding to gpstimes
        background = []
        segments = [
            tuple(map(float, f.name.split(".")[0].split("-")[1:]))
            for f in fnames
        ]

        def find_file(time: float) -> Optional[Path]:
            """
            Find file that contains `time`
            """
            for i, (start, length) in enumerate(segments):
                in_segment = time > start + self.sample_length
                in_segment &= time < (start + length - self.sample_length)
                if in_segment:
                    return self.test_fnames[i], start
            else:
                return None, None

        # calculate seconds of data to query
        # before and after coalescence time
        post = self.waveform_sampler.right_pad + self.hparams.fduration / 2
        pre = (
            post
            - self.hparams.kernel_length
            - (self.hparams.fduration / 2)
            - self.hparams.psd_length
        )

        # convert to number of indices
        post = int(post * self.hparams.sample_rate)
        pre = int(pre * self.hparams.sample_rate)

        background = []
        for time in gpstimes:
            time = time.item()
            strain = []

            # find file for this gpstime
            file, start = find_file(time)

            # if none exists, use random segment
            if file is None:
                self._logger.info(
                    "No segment in testing directory containing "
                    f"{time}. Using random segment"
                )
                file = random.choice(self.test_fnames)
                start, length = list(
                    map(float, file.name.split(".")[0].split("-")[1:])
                )
                time = start + random.randint(
                    self.sample_length,
                    length - self.sample_length,
                )

            # convert from time to index in file
            middle_idx = int((time - start) * self.hparams.sample_rate)
            start_idx = middle_idx + pre
            end_idx = middle_idx + post

            with h5py.File(file) as f:
                for ifo in self.hparams.ifos:
                    strain.append(torch.tensor(f[ifo][start_idx:end_idx]))
                strain = torch.stack(strain, dim=0)
                background.append(strain)
        background = torch.stack(background, dim=0)
        return background



import logging
from contextlib import nullcontext
from pathlib import Path
from typing import Optional, Sequence

import torch
import zuko

from .embeddings.base import Embedding


class FlowArchitecture(torch.nn.Module):
    """
    Base class for normalizing flow architectures that
    provides interface for interacting with embedding networks
    """

    def __init__(
        self,
        num_params: int,
        embedding_net: Embedding,
        embedding_weights: Optional[Path] = None,
        freeze_embedding: bool = False,
    ):
        super().__init__()
        self.num_params = num_params
        self.embedding_net = embedding_net

        if freeze_embedding:
            self.embedding_context = torch.no_grad
        else:
            self.embedding_context = nullcontext

        if embedding_weights is not None:
            # load in pre trained embedding weights,
            # removing extra module weights (like, e.g. the standard scaler)
            logging.info(f"Loading embedding weights from {embedding_weights}")
            checkpoint = torch.load(embedding_weights)
            state_dict = checkpoint["state_dict"]
            state_dict = {
                k.removeprefix("model.embedding"): v
                for k, v in state_dict.items()
                if k.startswith("model.embedding")
            }
            self.embedding_net.load_state_dict(state_dict)

    def build_flow(self) -> zuko.lazy.Flow:
        raise NotImplementedError

    def log_prob(self, x, context):
        """Wrapper around :meth:`log_prob` from
        `zuko.lazy.Flow` object.
        """
        if not hasattr(self, "flow"):
            raise RuntimeError("Flow is not built")

        with self.embedding_context():
            embedded_context = self.embedding_net(context)
        return self.flow(embedded_context).log_prob(x)

    def sample(self, n, context):
        """Wrapper around :meth:`sample` from
        `TransformedDistribution` object.
        """
        if not hasattr(self, "flow"):
            raise RuntimeError("Flow is not built")
        embedded_context = self.embedding_net(context)
        return self.flow(embedded_context).sample((n,))


class NSF(FlowArchitecture):
    """
    Light wrapper around the `NSF` flow from `zuko` library
    for compatibility with the `FlowArchitecture` interface.

    See https://zuko.readthedocs.io/stable/api/zuko.flows.spline.html#zuko.flows.spline.NSF

    Args:
        transforms:
            Number of transformations in the flow
        hidden_features:
            Sequence of integers representing hidden units in the hyper network
        passes:
            Default of `None` corresponds to fully autoregressive flow.
            A value of 2 corresponds to coupling flow.
        bins:
            Number of bins in the spline
        randperm:
            Whether to randomly permute features in between
            transformation layers
        residual:
            Whether to use residual connections in the hyper network.

    """  # noqa E501

    def __init__(
        self,
        *args,
        transforms: int,
        hidden_features: Optional[Sequence[int]] = (64, 64),
        passes: Optional[int] = None,
        bins: Optional[int] = 8,
        randperm: Optional[bool] = False,
        residual: Optional[bool] = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.transforms = transforms
        self.bins = bins
        self.randperm = randperm
        self.hidden_features = hidden_features
        self.residual = residual
        self.passes = passes
        self.flow = self.build_flow()

    def build_flow(self) -> zuko.lazy.Flow:
        return zuko.flows.NSF(
            self.num_params,
            transforms=self.transforms,
            context=self.embedding_net.context_dim,
            bins=self.bins,
            randperm=self.randperm,
            hidden_features=self.hidden_features,
            passes=self.passes,
            residual=self.residual,
        )



import os

import luigi
from mldatafind.law.parameters import PathParameter


class paths(luigi.Config):
    data_dir = PathParameter(default=os.getenv("AMPLFI_DATADIR", ""))
    condor_dir = PathParameter(default=os.getenv("AMPLFI_CONDORDIR", ""))
    container_root = PathParameter(
        default=os.getenv("AMPLFI_CONTAINER_ROOT", "")
    )



from typing import TYPE_CHECKING, Optional

import torch

from ...utils.utils import ParameterSampler
from ..sampler import WaveformSampler

if TYPE_CHECKING:
    from ml4gw.transforms import ChannelWiseScaler


class WaveformGenerator(WaveformSampler):
    def __init__(
        self,
        *args,
        num_val_waveforms: int,
        num_test_waveforms: int,
        parameter_sampler: ParameterSampler,
        test_parameter_sampler: Optional[ParameterSampler] = None,
        num_fit_params: int,
        **kwargs,
    ):
        """
        A torch module for generating waveforms on the fly.

        Args:
            num_val_waveforms:
                Total number of validation waveforms to use.
                This total will be split up among all devices
            num_test_waveforms:
                Total number of testing waveforms to use.
                Testing is performed on one device.
            parameter_sampler:
                A callable that takes an integer N and
                returns a dictionary of parameter Tensors, each of length `N`
            test_parameter_sampler:
                A callable that takes an integer N and
                returns a dictionary of parameter Tensors, each of length `N`.
                Used for sampling test waveforms from a prior
                different from training data.
                If None, `parameter_sampler` is used.
            num_fit_params:
                The number of parameters used to fit standard scaler

        """
        super().__init__(*args, **kwargs)
        self.parameter_sampler = parameter_sampler
        self.test_parameter_sampler = (
            test_parameter_sampler or parameter_sampler
        )
        self.num_val_waveforms = num_val_waveforms
        self.num_test_waveforms = num_test_waveforms
        self.num_fit_params = num_fit_params

    def get_val_waveforms(self, _, world_size):
        num_waveforms = self.num_val_waveforms // world_size
        parameters = self.parameter_sampler(num_waveforms, device="cpu")
        hc, hp = self(**parameters)
        return hc, hp, parameters

    def get_test_waveforms(self):
        parameters = self.test_parameter_sampler(self.num_test_waveforms)
        hc, hp = self(**parameters)
        return hc, hp, parameters

    def sample(self, X):
        N = len(X)
        parameters = self.parameter_sampler(N, device=X.device)
        hc, hp = self(**parameters)
        return hc, hp, parameters

    def fit_scaler(self, scaler: "ChannelWiseScaler") -> "ChannelWiseScaler":
        parameters = self.parameter_sampler(self.num_fit_params)

        dec, psi, phi = self.sample_extrinsic(torch.ones(self.num_fit_params))
        parameters.update({"dec": dec, "psi": psi, "phi": phi})
        transformed = self.parameter_transformer(parameters)

        fit = []
        for key in self.inference_params:
            fit.append(transformed[key])

        fit = torch.row_stack(fit)
        scaler.fit(fit)
        return scaler

    def forward(self):
        raise NotImplementedError



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import sys
from importlib.metadata import distribution
from pathlib import Path

parent_path = str(Path(__file__).parents[1])
sys.path.insert(0, parent_path)


# -- Project information -----------------------------------------------------
dist = distribution("amplfi")

project = dist.metadata["Name"]
author = dist.metadata["Author"]
copyright = f"2024, {author}"
release = dist.metadata["Version"]

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.extlinks",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_tabs.tabs",
    "sphinx_autodoc_typehints",
    "sphinx.ext.githubpages",
    "myst_parser",
]

extlinks = {
    "doi": ("https://dx.doi.org/%s", "doi:%s"),
}

myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "strikethrough",
    "substitution",
    "tasklist",
]

source_suffix = [".rst", ".md"]


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "python"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

autodoc_default_options = {
    "members": None,
    "undoc-members": None,
    "show-inheritance": None,
}
autodoc_inherit_docstrings = False
autodoc_member_order = "bysource"
autosummary_generate = True

modindex_common_prefix = ["amplfi."]



import pytest
import torch

from amplfi.train.architectures.embeddings import MultiModal, ResNet
from amplfi.train.architectures.embeddings.dense import DenseEmbedding


@pytest.fixture(params=[1, 2, 3])
def n_ifos(request):
    return request.param


@pytest.fixture(params=[2048, 8192])
def length(request):
    return request.param


@pytest.fixture(
    params=[
        3,
        5,
    ]
)
def kernel_size(request):
    return request.param


@pytest.fixture(params=[128, 256])
def out_features(request):
    return request.param


@pytest.fixture(params=[32, 64, 128])
def time_out_features(request):
    return request.param


@pytest.fixture(params=[32, 64, 128])
def freq_out_features(request):
    return request.param


def test_dense_embedding(n_ifos, length):
    embedding = DenseEmbedding(length, 10)
    x = torch.randn(8, n_ifos, length)
    y = embedding(x)

    assert y.shape == (8, n_ifos, 10)


def test_resnet(n_ifos, length, out_features, kernel_size):
    embedding = ResNet(n_ifos, out_features, [3, 3], kernel_size)
    x = torch.randn(100, n_ifos, length)
    y = embedding((x, None))
    assert y.shape == (100, out_features)


def test_multimodal(
    n_ifos, length, time_out_features, freq_out_features, kernel_size
):
    embedding = MultiModal(
        n_ifos,
        time_out_features,
        freq_out_features,
        [3, 3],
        [3, 3],
        kernel_size,
        kernel_size,
    )
    x = (torch.randn(100, n_ifos, length), None)
    y = embedding(x)
    assert y.shape == (100, time_out_features + freq_out_features)



from lightning.pytorch.cli import LightningCLI

from ..callbacks import SaveConfigCallback


class AmplfiBaseCLI(LightningCLI):
    def __init__(self, *args, **kwargs):
        # hack into init to hardcode
        # parser_mode to omegaconf for all subclasses
        kwargs["parser_kwargs"] = {"parser_mode": "omegaconf"}
        kwargs["save_config_callback"] = SaveConfigCallback
        kwargs["save_config_kwargs"] = {"overwrite": True}
        super().__init__(*args, **kwargs)

    def add_arguments_to_parser(self, parser):
        parser.link_arguments(
            "data.init_args.sample_rate",
            "data.init_args.waveform_sampler.init_args.sample_rate",
            apply_on="parse",
        )

        parser.link_arguments(
            "data.init_args.kernel_length",
            "data.init_args.waveform_sampler.init_args.kernel_length",
            apply_on="parse",
        )

        parser.link_arguments(
            "data.init_args.fduration",
            "data.init_args.waveform_sampler.init_args.fduration",
            apply_on="parse",
        )

        parser.link_arguments(
            "data.init_args.inference_params",
            "data.init_args.waveform_sampler.init_args.inference_params",
            apply_on="parse",
        )

        parser.link_arguments(
            "data.init_args.inference_params",
            "model.init_args.inference_params",
            apply_on="parse",
        )



import bilby
from typing import TYPE_CHECKING
import numpy as np
import pandas as pd
import torch

from ..architectures.flows import FlowArchitecture
from ..callbacks import (
    StrainVisualization,
    SavePosterior,
    CrossMatchStatistics,
    ProbProbPlot,
)
from ...utils.result import AmplfiResult
from .base import AmplfiModel
from typing import Optional

if TYPE_CHECKING:
    pass

Tensor = torch.Tensor


class FlowModel(AmplfiModel):
    """
    A LightningModule for training normalizing flows

    Args:
        *args:
            See arguments in `amplfi.train.models.base.AmplfiModel`
        arch:
            Neural network architecture to train.
            This should be a subclass of `FlowArchitecture`.
        samples_per_event:
            Number of samples to draw per event for testing
        nside:
            Healpix nside used when creating skymaps
        min_samples_per_pix:
        num_plot:
            Number of testing events to plot skymaps, corner
            plots and, if `plot_data` is `True`, strain data
            visualizations.
        plot_data:
            If `True`, plot strain visualization for `num_plot`
            of the testing set events
        save_posterior:
            If `True`, save bilby Result objects and posterior samples
        cross_match:
            If `True`, run ligo.skymap.postprocess.crossmatch
            on result objects at the end of testing epoch
            and produce searched area and volume cdfs
    """

    def __init__(
        self,
        *args,
        arch: FlowArchitecture,
        samples_per_event: int = 10000,
        nside: int = 32,
        min_samples_per_pix: int = 15,
        num_plot: int = 10,
        plot_data: bool = False,
        save_posterior: bool = False,
        cross_match: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.model = arch
        self.samples_per_event = samples_per_event
        self.num_plot = num_plot
        self.nside = nside
        self.min_samples_per_pix = min_samples_per_pix
        self.plot_data = plot_data
        self.save_posterior = save_posterior
        self.cross_match = cross_match

        # save our hyperparameters
        self.save_hyperparameters(ignore=["arch"])

    def forward(self, context, parameters) -> Tensor:
        return -self.model.log_prob(parameters, context=context)

    def training_step(self, batch, _):
        strain, asds, parameters = batch
        context = (strain, asds)
        loss = self(context, parameters).mean()
        self.log(
            "train_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            sync_dist=True,
            logger=True,
        )
        return loss

    def validation_step(self, batch, _):
        strain, asds, parameters = batch
        context = (strain, asds)
        loss = self(context, parameters).mean()
        self.log(
            "valid_loss",
            loss,
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            sync_dist=True,
            logger=True,
        )
        return loss

    def test_step(self, batch, batch_idx) -> AmplfiResult:
        strain, asds, parameters = batch
        context = (strain, asds)

        samples = self.model.sample(
            self.hparams.samples_per_event, context=context
        )
        samples = samples.squeeze(1)
        descaled = self.scale(samples, reverse=True)
        descaled = self.filter_parameters(descaled)
        parameters = self.scale(parameters, reverse=True)

        result = self.cast_as_bilby_result(
            descaled.cpu().numpy(),
            parameters.cpu().numpy()[0],
        )

        test_outdir = self.test_outdir / f"event_{batch_idx}"
        test_outdir.mkdir(parents=True, exist_ok=True)

        # add ra column for use with ligo-skymap-from-samples
        result.posterior["ra"] = result.posterior["phi"]

        self.test_results.append(result)

        # plot corner and skymap
        skymap_filename = test_outdir / "mollview.png"
        corner_filename = test_outdir / "corner.png"
        # fits_filename = test_outdir / "amplfi.skymap.fits"
        result.plot_corner(
            save=True,
            filename=corner_filename,
            levels=(0.5, 0.9),
        )
        result.plot_mollview(
            outpath=skymap_filename,
        )

        return result

    def predict_step(self, batch, batch_idx, _):
        strain, asds, gpstime = batch
        context = (strain, asds)

        samples = self.model.sample(
            self.hparams.samples_per_event, context=context
        )
        descaled = self.scale(samples, reverse=True)
        descaled = self.filter_parameters(descaled)
        result = self.cast_as_bilby_result(
            descaled.cpu().numpy(),
            None,
        )

        test_outdir = self.test_outdir / f"event_{batch_idx}"
        test_outdir.mkdir(parents=True, exist_ok=True)

        skymap_filename = test_outdir / f"{gpstime.item()}_mollview.png"
        corner_filename = test_outdir / f"{gpstime.item()}_corner.png"
        # fits_filename = test_outdir / f"{gpstime.item()}.fits"

        result.plot_corner(
            save=True,
            filename=corner_filename,
            levels=(0.5, 0.9),
        )
        result.plot_mollview(
            outpath=skymap_filename,
        )

        return result

    def on_test_epoch_start(self):
        self.test_results: list[AmplfiResult] = []

    def cast_as_bilby_result(
        self,
        samples: np.ndarray,
        truth: Optional[np.ndarray] = None,
    ) -> AmplfiResult:
        """Cast posterior samples as Bilby Result object
        for ease of producing corner and pp plots

        Args:
            samples: posterior samples (1, num_samples, num_params)
            truth: true values of the parameters  (1, num_params)

        """

        injection_parameters = (
            {
                k: float(v)
                for k, v in zip(self.inference_params, truth, strict=False)
            }
            if truth is not None
            else None
        )

        # create dummy prior with correct attributes
        # for making our results compatible with bilbys make_pp_plot
        priors = {
            param: bilby.core.prior.base.Prior(latex_label=param)
            for param in self.inference_params
        }
        posterior = {}
        for idx, k in enumerate(self.inference_params):
            posterior[k] = samples.T[idx].flatten()
        posterior = pd.DataFrame(posterior)

        r = AmplfiResult(
            label="PEModel",
            injection_parameters=injection_parameters,
            posterior=posterior,
            search_parameter_keys=self.inference_params,
            priors=priors,
        )
        return r

    def filter_parameters(self, parameters: torch.Tensor):
        """
        Filter the descaled parameters to keep only valid samples
        within their boundaries.

        Args:
            descaled (torch.Tensor): The descaled parameters tensor.

        Returns:
            torch.Tensor: The filtered descaled parameters.
        """
        net_mask = torch.ones(
            parameters.shape[0], dtype=bool, device=parameters.device
        )
        waveform_sampler = self.trainer.datamodule.waveform_sampler
        priors = waveform_sampler.parameter_sampler.parameters
        for i, param in enumerate(self.inference_params):
            samples = parameters[:, i]
            if param in ["dec", "phi", "psi"]:
                prior = getattr(
                    self.trainer.datamodule.waveform_sampler, param
                )
            else:
                prior = priors[param]

            mask = (prior.log_prob(samples) == float("-inf")).to(
                samples.device
            )
            self._logger.debug(
                f"Removed {mask.sum()}/{len(mask)} samples for parameter "
                f"{param} outside of prior range"
            )

            net_mask &= ~mask

        self._logger.info(
            f"Removed {(~net_mask).sum()}/{len(net_mask)} total samples "
            "outside of prior range"
        )
        parameters = parameters[net_mask]

        return parameters

    def configure_callbacks(self):
        callbacks = [ProbProbPlot()]
        if self.plot_data:
            callbacks.append(
                StrainVisualization(self.test_outdir, self.num_plot)
            )

        if self.save_posterior:
            callbacks.append(SavePosterior(self.test_outdir))

        if self.cross_match:
            callbacks.append(CrossMatchStatistics())

        return callbacks



from typing import Literal, Optional

import torch
from ml4gw.nn.norm import NormLayer
from ml4gw.nn.resnet.resnet_1d import ResNet1D

from .base import Embedding


class MultiModal(Embedding):
    def __init__(
        self,
        num_ifos: int,
        time_context_dim: int,
        freq_context_dim: int,
        time_layers: list[int],
        freq_layers: list[int],
        time_kernel_size: int = 3,
        freq_kernel_size: int = 3,
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group: int = 64,
        stride_type: Optional[list[Literal["stride", "dilation"]]] = None,
        norm_layer: Optional[NormLayer] = None,
        **kwargs,
    ):
        """
        MultiModal embedding network that embeds both time and frequency data.

        We pass the data through their own ResNets defined by their layers
        and context dims, then concatenate the output embeddings.
        """
        super().__init__()
        self.time_domain_resnet = ResNet1D(
            in_channels=num_ifos,
            layers=time_layers,
            classes=time_context_dim,
            kernel_size=time_kernel_size,
            zero_init_residual=zero_init_residual,
            groups=groups,
            width_per_group=width_per_group,
            stride_type=stride_type,
            norm_layer=norm_layer,
        )
        self.frequency_domain_resnet = ResNet1D(
            in_channels=int(num_ifos * 2),
            layers=freq_layers,
            classes=freq_context_dim,
            kernel_size=freq_kernel_size,
            zero_init_residual=zero_init_residual,
            groups=groups,
            width_per_group=width_per_group,
            stride_type=stride_type,
            norm_layer=norm_layer,
        )

        # set the context dimension so
        # the flow can access it
        self.context_dim = time_context_dim + freq_context_dim

    def forward(self, X):
        # unpack, ignoring asds
        strain, _ = X
        time_domain_embedded = self.time_domain_resnet(strain)
        strain_fft = torch.fft.rfft(strain)
        strain_fft = torch.cat((strain_fft.real, strain_fft.imag), dim=1)
        frequency_domain_embedded = self.frequency_domain_resnet(strain_fft)

        embedding = torch.concat(
            (time_domain_embedded, frequency_domain_embedded), dim=1
        )
        return embedding


class MultiModalPsd(Embedding):
    """
    MultiModal embedding network that embeds both time and frequency data.

    We pass the data through their own ResNets defined by their layers
    and context dims, then concatenate the output embeddings.
    """

    def __init__(
        self,
        num_ifos: int,
        time_context_dim: int,
        freq_context_dim: int,
        time_layers: list[int],
        freq_layers: list[int],
        time_kernel_size: int = 3,
        freq_kernel_size: int = 3,
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group: int = 64,
        stride_type: Optional[list[Literal["stride", "dilation"]]] = None,
        norm_layer: Optional[NormLayer] = None,
        **kwargs,
    ):
        super().__init__()
        self.context_dim = time_context_dim + freq_context_dim
        self.time_domain_resnet = ResNet1D(
            in_channels=num_ifos,
            layers=time_layers,
            classes=time_context_dim,
            kernel_size=time_kernel_size,
            zero_init_residual=zero_init_residual,
            groups=groups,
            width_per_group=width_per_group,
            stride_type=stride_type,
            norm_layer=norm_layer,
        )

        self.freq_psd_resnet = ResNet1D(
            in_channels=int(num_ifos * 3),
            layers=freq_layers,
            classes=freq_context_dim,
            kernel_size=freq_kernel_size,
            zero_init_residual=zero_init_residual,
            groups=groups,
            width_per_group=width_per_group,
            stride_type=stride_type,
            norm_layer=norm_layer,
        )

    def forward(self, X):
        strain, asds = X

        asds *= 1e23
        asds = asds.float()
        inv_asds = 1 / asds

        time_domain_embedded = self.time_domain_resnet(strain)
        X_fft = torch.fft.rfft(strain)
        X_fft = X_fft[..., -asds.shape[-1] :]
        X_fft = torch.cat((X_fft.real, X_fft.imag, inv_asds), dim=1)

        frequency_domain_embedded = self.freq_psd_resnet(X_fft)
        embedding = torch.concat(
            (time_domain_embedded, frequency_domain_embedded),
            dim=1,
        )
        return embedding


class FrequencyPsd(Embedding):
    """
    Single embedding for frequency domain data with ASDS
    """

    def __init__(
        self,
        num_ifos: int,
        context_dim: int,
        layers: list[int],
        kernel_size: int = 3,
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group: int = 64,
        stride_type: Optional[list[Literal["stride", "dilation"]]] = None,
        norm_layer: Optional[NormLayer] = None,
    ):
        super().__init__()
        self.freq_psd_resnet = ResNet1D(
            in_channels=int(num_ifos * 3),
            layers=layers,
            classes=context_dim,
            kernel_size=kernel_size,
            zero_init_residual=zero_init_residual,
            groups=groups,
            width_per_group=width_per_group,
            stride_type=stride_type,
            norm_layer=norm_layer,
        )
        self.context_dim = context_dim

    def forward(self, X):
        strain, asds = X

        asds *= 1e23
        asds = asds.float()
        inv_asds = 1 / asds

        X_fft = torch.fft.rfft(strain)
        X_fft = X_fft[..., -asds.shape[-1] :]
        X_fft = torch.cat((X_fft.real, X_fft.imag, inv_asds), dim=1)
        embedding = self.freq_psd_resnet(X_fft)

        return embedding



#!/usr/bin/env python3

import logging
import os
import shutil
from pathlib import Path
from textwrap import dedent
from typing import Literal, Optional

import yaml
from jsonargparse import ArgumentParser

root = Path(__file__).resolve().parent.parent
data_config = (root / "amplfi" / "data" / "datagen.cfg",)
kubernetes_config = root / "kubernetes" / "train.yaml"
TUNE_CONFIGS = [
    root / "amplfi" / "train" / "configs" / "tune.yaml",
]


TRAIN_CONFIGS = {
    "similarity": [
        root / "amplfi" / "train" / "configs" / "similarity" / "cbc.yaml"
    ],
    "flow": [root / "amplfi" / "train" / "configs" / "flow" / "cbc.yaml"],
}


def fill_kubernetes_template(output: Path, s3_bucket):
    """
    Fill in the kubernetes template with the users environment variables
    """
    with open(kubernetes_config) as f:
        docs = list(yaml.safe_load_all(f))
        config, s3 = docs
        s3["stringData"]["AWS_ACCESS_KEY_ID"] = os.getenv(
            "AWS_ACCESS_KEY_ID", ""
        )
        s3["stringData"]["AWS_SECRET_ACCESS_KEY"] = os.getenv(
            "AWS_SECRET_ACCESS_KEY", ""
        )

        # set remote training config path
        config["spec"]["template"]["spec"]["containers"][0]["args"][2] = (
            f"{s3_bucket}/cbc.yaml"
        )

        # set environment variables that will
        # be used in the training job by lightning
        config["spec"]["template"]["spec"]["containers"][0]["env"][1][
            "value"
        ] = os.getenv("WANDB_API_KEY", "")
        config["spec"]["template"]["spec"]["containers"][0]["env"][2][
            "value"
        ] = s3_bucket
        config["spec"]["template"]["spec"]["containers"][0]["env"][3][
            "value"
        ] = f"{s3_bucket}/data"

        with open(output, "w") as f:
            yaml.safe_dump_all([config, s3], f)


def copy_configs(
    path: Path,
    configs: list[Path],
):
    """
    Copy the configuration files to the specified directory for editing.

    Any path specific configurations will be updated to point to the
    correct paths in the new directory.

    Args:
        path:
            The directory to copy the configuration files to.
        configs:
            The list of configuration files to copy.
    """

    path.mkdir(parents=True, exist_ok=True)
    for config in configs:
        dest = path / config.name
        shutil.copy(config, dest)


def write_content(content: str, path: Path):
    content = dedent(content).strip("\n")
    with open(path, "w") as f:
        f.write(content)

    # make the file executable
    path.chmod(0o755)
    return content


def create_remote_runfile(
    path: Path,
    name: str,
    s3_bucket: Path,
):
    rundir = path / name
    config = rundir / "datagen.cfg"
    content = f"""
    #!/bin/bash
    export AMPLFI_DATADIR={s3_bucket}/data

    # launch data generation pipeline
    LAW_CONFIG_FILE={config} law run amplfi.data.DataGeneration --workers 5

    # move config file to remote s3 location
    s3cmd put {rundir}/cbc.yaml {s3_bucket}/cbc.yaml

    # launch job
    kubectl apply -f {rundir}/kubernetes.yaml

    """

    runfile = path / name / "run.sh"
    write_content(content, runfile)


def create_runfile(
    path: Path,
    name: str,
    mode: Literal["flow", "similarity"],
    pipeline: Literal["tune", "train"],
    s3_bucket: Optional[Path] = None,
):
    # if s3 bucket is provided
    # store training data and training info there
    base = path if s3_bucket is None else s3_bucket

    config = path / name / "datagen.cfg"
    # make the below one string
    data_cmd = f"LAW_CONFIG_FILE={config} "
    data_cmd += "law run amplfi.data.DataGeneration --workers 5"

    if pipeline == "tune":
        train_cmd = "lightray --config tune.yaml -- --config cbc.yaml"
    else:
        train_cmd = f"amplfi-{mode}-cli fit --config cbc.yaml"

    content = f"""
    #!/bin/bash
    # set environment variables for this job
    export AMPLFI_DATADIR={base}/data/
    export AMPLFI_OUTDIR={base}/{name}/
    export AMPLFI_CONDORDIR={path}/data/condor

    # set the GPUs exposed to job
    CUDA_VISIBLE_DEVICES=0

    # launch the data generation pipeline
    {data_cmd}

    # launch {pipeline} pipeline
    {train_cmd}
    """

    runfile = path / name / "run.sh"
    write_content(content, runfile)


def main():
    parser = ArgumentParser(
        description="Initialize a directory with configuration files "
        "for running end-to-end amplfi training or tuning pipelines"
    )
    parser.add_argument(
        "--mode",
        choices=["flow", "similarity"],
        default="flow",
        help="Either 'flow' or 'similarity'. "
        "Whether to setup a flow or similarity training",
    )
    parser.add_argument(
        "--pipeline",
        choices=["tune", "train"],
        default="train",
        help="Either 'train' or 'tune'. "
        "Whether to setup a tune or train pipeline",
    )
    parser.add_argument(
        "-n",
        "--name",
        type=str,
        required=True,
        help="The name of the run. "
        "This will be used to create the run subdirectory.",
    )
    parser.add_argument(
        "-d",
        "--directory",
        type=Path,
        default=None,
        help="The parent directory where the "
        "data and subdirectories for runs will "
        "be stored. If not provided, the environment "
        "variable AMPLFI_RUNDIR will be used.",
    )

    parser.add_argument(
        "--remote-train",
        type=bool,
        default=None,
        help="Whether to train remotely on nautilus. "
        "If `True`, will copy a yaml file to the run directory "
        "that contains kubernetes yaml for a remote training job.",
    )

    parser.add_argument(
        "--s3-bucket",
        type=str,
        default=None,
        help="The s3 bucket where training data is stored",
    )
    log_format = "%(levelname)s - %(message)s"
    logging.basicConfig(level=logging.INFO, format=log_format)
    args = parser.parse_args()
    directory = (
        args.directory.resolve()
        if args.directory
        else Path(os.environ.get("AMPLFI_RUNDIR")).resolve()
    )

    if args.s3_bucket is not None:
        s3_bucket = args.s3_bucket.rstrip("/")
        if not args.s3_bucket.startswith("s3://"):
            raise ValueError(
                "S3 bucket must be in the format s3://{bucket-name}/"
            )
    elif args.remote_train and args.s3_bucket is None:
        raise ValueError(
            "S3 bucket must be provided to train remotely on nautilus"
        )
    else:
        s3_bucket = None

    # construct the config files to copy
    # for the given mode and pipeline
    if args.pipeline == "tune":
        configs = TUNE_CONFIGS
        configs.extend(data_config)
        configs.extend(TRAIN_CONFIGS[args.mode])
    else:
        configs = TRAIN_CONFIGS[args.mode]
        configs.extend(data_config)

    copy_configs(directory / args.name, configs)

    if args.remote_train:
        fill_kubernetes_template(
            directory / args.name / "kubernetes.yaml", s3_bucket
        )
        create_remote_runfile(directory, args.name, s3_bucket)
    else:
        create_runfile(
            directory, args.name, args.mode, args.pipeline, s3_bucket
        )

    logging.info(
        f"Initialized a {args.mode} {args.pipeline} "
        f"pipeline at {directory / args.name}"
    )


if __name__ == "__main__":
    main()



"""Auxiliary functions for distance ansatz see:10.3847/2041-8205/829/1/L15"""

import numpy as np
import scipy as sp


def P(x):
    return np.exp(-0.5 * x**2) / np.sqrt(2 * np.pi)


def Q(x):
    return sp.special.erfc(x / np.sqrt(2)) / 2


def H(x):
    return P(x) / Q(x)


def dHdz(z):
    return -H(z) * (H(z) - z)


def x2(z):
    return z**2 + 1 + z * H(-z)


def x3(z):
    return z**3 + 3 * z + (z**2 + 2) * H(-z)


def x4(z):
    return z**4 + 6 * z**2 + 3 + (z**3 + 5 * z) * H(-z)


def x2_prime(z):
    return 2 * z + H(-z) + z * dHdz(-z)


def x3_prime(z):
    return 3 * z**2 + 3 + 2 * z * H(-z) + (z**2 + 2) * dHdz(-z)


def x4_prime(z):
    return (
        4 * z**3 + 12 * z + (3 * z**2 + 5) * H(-z) + (z**3 + 5 * z) * dHdz(-z)
    )


def f(z, s, m):
    r = 1 + (s / m) ** 2
    r *= x3(z) ** 2
    r -= x2(z) * x4(z)
    return r


def fprime(z, s, m):
    r = 2 * (1 + (s / m) ** 2)
    r *= x3(z) * x3_prime(z)
    r -= x2(z) * x4_prime(z)
    r -= x2_prime(z) * x4(z)
    return r


def moments_from_samples_impl(d):
    """
    Given distance samples, assumed to be posterior
    samples, compute distance moments in monte-carlo sense.

    Args:
        d:
           Distance samples
    """
    d_2 = d**2
    rho = d_2.sum()
    d_3 = d**3
    d_3 = d_3.sum()
    d_4 = d**4
    d_4 = d_4.sum()

    m = d_3 / rho
    s = np.sqrt(d_4 / rho - m**2)
    return rho, m, s


def ansatz_impl(s, m, maxiter=10):
    """Given s and m parameters (see Eqs. 1-3 of
    :doi:`10.3847/0067-0049/226/1/10`) solve for
    mu, sigma, and norm parameters.

    Args:
        s:
            Std. dev calculated from moments
        m:
            Conditional distance mean per pixel
    """
    z0 = m / s
    sol = sp.optimize.root_scalar(
        f, args=(s, m), fprime=fprime, x0=z0, maxiter=maxiter
    )
    if not sol.converged:
        dist_mu = float("inf")
        dist_sigma = 1
        dist_norm = 0
    else:
        z_hat = sol.root
        dist_sigma = m * x2(z_hat) / x3(z_hat)
        dist_mu = dist_sigma * z_hat
        dist_norm = 1 / (Q(-z_hat) * dist_sigma**2 * x2(z_hat))
    return dist_mu, dist_sigma, dist_norm



from typing import Literal, Optional

from ml4gw.nn.norm import NormLayer
from ml4gw.nn.resnet.resnet_1d import ResNet1D

from .base import Embedding


class ResNet(ResNet1D, Embedding):
    def __init__(
        self,
        num_ifos: int,
        context_dim: int,
        layers: list[int],
        kernel_size: int = 3,
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group: int = 64,
        stride_type: Optional[list[Literal["stride", "dilation"]]] = None,
        norm_layer: Optional[NormLayer] = None,
    ):
        super().__init__(
            num_ifos,
            layers=layers,
            classes=context_dim,
            kernel_size=kernel_size,
            zero_init_residual=zero_init_residual,
            groups=groups,
            width_per_group=width_per_group,
            stride_type=stride_type,
            norm_layer=norm_layer,
        )

        # set the context dimension so
        # the flow can access it
        self.context_dim = context_dim

    def forward(self, x):
        strain, _ = x
        return super().forward(strain)



import torch
import torch.nn.functional as F
from torch import Tensor


class VICRegLoss(torch.nn.Module):
    """Implementation of the VICReg loss [0].

    This implementation is based on the code published by the authors [1].

    - [0] VICReg, 2022, https://arxiv.org/abs/2105.04906
    - [1] https://github.com/facebookresearch/vicreg/

    Attributes:
        lambda_param:
            Scaling coefficient for the invariance term of the loss.
        mu_param:
            Scaling coefficient for the variance term of the loss.
        nu_param:
            Scaling coefficient for the covariance term of the loss.
        eps:
            Epsilon for numerical stability.

    """

    def __init__(
        self,
        lambda_param: float = 25.0,
        mu_param: float = 25.0,
        nu_param: float = 1.0,
        eps: float = 0.0001,
        max_std: float = 1.0,
    ):
        """Initializes the VICRegLoss module with the specified parameters."""
        super(VICRegLoss, self).__init__()

        self.lambda_param = lambda_param
        self.mu_param = mu_param
        self.nu_param = nu_param
        self.max_std = max_std
        self.eps = eps

    def forward(self, z_a: torch.Tensor, z_b: torch.Tensor) -> torch.Tensor:
        """Returns VICReg loss.

        Args:
            z_a:
                Tensor with shape (batch_size, ..., dim).
            z_b:
                Tensor with shape (batch_size, ..., dim).

        Returns:
            The computed VICReg loss.

        """

        # Invariance term of the loss
        inv_loss = invariance_loss(x=z_a, y=z_b)

        # Variance and covariance terms of the loss
        var_loss = 0.5 * (
            variance_loss(x=z_a, eps=self.eps, max_std=self.max_std)
            + variance_loss(x=z_b, eps=self.eps, max_std=self.max_std)
        )
        cov_loss = covariance_loss(x=z_a) + covariance_loss(x=z_b)

        # Total VICReg loss
        loss = (
            self.lambda_param * inv_loss
            + self.mu_param * var_loss
            + self.nu_param * cov_loss
        )
        return loss, (inv_loss, var_loss, cov_loss)


def invariance_loss(x: Tensor, y: Tensor) -> Tensor:
    """Returns VICReg invariance loss.

    Args:
        x:
            Tensor with shape (batch_size, ..., dim).
        y:
            Tensor with shape (batch_size, ..., dim).

    Returns:
        The computed VICReg invariance loss.
    """
    return F.mse_loss(x, y)


def variance_loss(
    x: Tensor, eps: float = 0.0001, max_std: float = 1.0
) -> Tensor:
    """Returns VICReg variance loss.

    Args:
        x:
            Tensor with shape (batch_size, ..., dim).
        eps:
            Epsilon for numerical stability.

    Returns:
        The computed VICReg variance loss.
    """
    std = torch.sqrt(x.var(dim=0) + eps)
    loss = torch.mean(F.relu(max_std - std))
    return loss


def covariance_loss(x: Tensor) -> Tensor:
    """Returns VICReg covariance loss.

    Generalized version of the covariance loss with support for tensors with more than
    two dimensions. Adapted from VICRegL:
    https://github.com/facebookresearch/VICRegL/blob/803ae4c8cd1649a820f03afb4793763e95317620/main_vicregl.py#L299

    Args:
        x: Tensor with shape (batch_size, ..., dim).

    Returns:
          The computed VICReg covariance loss.
    """  # noqa
    x = x - x.mean(dim=0)
    batch_size = x.size(0)
    dim = x.size(-1)
    # nondiag_mask has shape (dim, dim) with 1s on all non-diagonal entries.
    nondiag_mask = ~torch.eye(dim, device=x.device, dtype=torch.bool)

    # cov has shape (..., dim, dim)
    cov = torch.einsum("b...c,b...d->...cd", x, x) / (batch_size - 1)

    loss = cov[..., nondiag_mask].pow(2).sum(-1) / dim
    return loss.mean()



"""
Additional Lightning DataModules for testing
amplfi models across various usecases
"""

from pathlib import Path
from typing import Optional, Union

import h5py
import numpy as np
import torch
from astropy.time import Time
from ml4gw.waveforms.conversion import chirp_mass_and_mass_ratio_to_components

from amplfi.train.data.datasets import FlowDataset

from ..utils.utils import ZippedDataset


def phi_from_ra(ra: np.ndarray, gpstimes: np.ndarray) -> float:
    # get the sidereal time at the observation time
    gmsts = []
    for t in gpstimes:
        t = Time(t, format="gps", scale="utc")
        gmst = t.sidereal_time("mean", "greenwich").to("rad").value
        gmsts.append(gmst)

    gmsts = np.array(gmsts)
    # calculate the relative azimuthal angle in the range [0, 2pi]
    phi = np.remainder(ra - gmsts, 2 * np.pi)

    return phi


class StrainTestingDataset(FlowDataset):
    """
    Testing dataset for analyzing pre-made injections.

    Subclass of `amplfi.train.data.datasets.flow` and
    thus `amplfi.train.data.datasets.base`. See those classes
    for additional arguments and keyword arguments

    Args:
        dataset_path:
            Path to hdf5 file containing premade injections.
            For each interferometer being analyzed, the strain
            data should be stored in an hdf5 dataset named
            after that interferometer.The dataset should be
            of shape (batch, time). It is assumed that the coalescence
            time of the injection is placed in the middle of each sample
            of the array. In addition, the parameters used to generate
            each injection should live in a dataset named after the
            parameter, e.g. `chirp_mass`.


    """

    def __init__(self, dataset_path: Path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dataset_path = dataset_path
        self.i = 0

    def setup(self, stage):
        world_size, rank = self.get_world_size_and_rank()
        self._logger = self.get_logger(world_size, rank)
        if stage != "test":
            raise ValueError(
                "StrainTestingDataset should only be used for testing"
            )

        # load in the strain data and parameters
        strain = []
        parameters = {}
        with h5py.File(self.dataset_path, "r") as f:
            for ifo in self.hparams.ifos:
                strain.append(torch.tensor(f[ifo][:], dtype=torch.float32))

            for parameter in self.hparams.inference_params + ["ra", "gpstime"]:
                # skip phi since these are proper injections
                # where we'll need to convert ra to phi
                # given the gpstime
                if parameter == "phi":
                    continue

                parameters[parameter] = torch.tensor(
                    f[parameter][:], dtype=torch.float32
                )

        # convert ra to phi
        parameters["phi"] = torch.tensor(
            phi_from_ra(
                parameters["ra"].numpy(), parameters["gpstime"].numpy()
            )
        )
        strain = torch.stack(strain, dim=1)

        # based on psd length, fduration and kernel length, and padding,
        # determine slice indices. It is assumed the coalescence
        # time of the waveform is in the middle
        middle = strain.shape[-1] // 2
        post = self.waveform_sampler.right_pad + self.hparams.fduration / 2
        pre = (
            post
            - self.hparams.kernel_length
            - (self.hparams.fduration / 2)
            - self.hparams.psd_length
        )

        post = int(post * self.hparams.sample_rate)
        pre = int(pre * self.hparams.sample_rate)

        start, stop = middle + pre, middle + post
        strain = strain[..., start:stop]
        # convert parameters to a tensor
        parameters = [
            torch.Tensor(parameters[k]) for k in self.hparams.inference_params
        ]
        parameters = torch.vstack(parameters).T

        # once we've generated validation/testing waveforms on cpu,
        # build data augmentation modules
        # and transfer them to appropiate device
        self.build_transforms(stage)
        self.transforms_to_device()

        self.parameters = parameters
        self.strain = strain

    def test_dataloader(self) -> torch.utils.data.DataLoader:
        # build dataset and dataloader that will
        # simply load one injection (and its parameters) at a time
        dataset = torch.utils.data.TensorDataset(self.strain, self.parameters)

        return torch.utils.data.DataLoader(
            dataset, batch_size=1, num_workers=12, shuffle=False
        )

    def on_after_batch_transfer(self, batch, _):
        """
        Override of the on_after_batch_transfer hook
        defined in `BaseDataset`.

        This is necessary since we're we don't want to
        do any injections (they are already in the data)
        """

        if not self.trainer.testing:
            raise ValueError(
                "Use of the StrainTestingDataset is for testing only"
            )

        X, parameters = batch

        X, psds = self.psd_estimator(X)
        X = self.whitener(X, psds)

        # scale parameters
        parameters = self.scale(parameters)

        # calculate asds, interpolating to size of
        # frequency array so we can concatenate them
        # into a single tensor for the embedding
        freqs = torch.fft.rfftfreq(X.shape[-1], d=1 / self.hparams.sample_rate)
        num_freqs = len(freqs)
        psds = torch.nn.functional.interpolate(
            psds, size=(num_freqs,), mode="linear"
        )

        mask = freqs > self.hparams.highpass
        psds = psds[:, :, mask]
        asds = torch.sqrt(psds)

        return X, asds, parameters


class ParameterTestingDataset(FlowDataset):
    """
    Testing dataset for using `amplfi` on the fly generationg
    to create injections from a fixed, premade set of parameters

    Subclass of `amplfi.train.data.datasets.flow` and
    thus `amplfi.train.data.datasets.base`. See those classes
    for additional arguments and keyword arguments.

    Args:
        dataset_path:
            Path to hdf5 dataset containing parameters
            to inject. Should contain hdf5 datasets
            for each of the waveform generation parameters
        waveform_generation_parameters:
            An optional list of parameters that is used for generating
            waveforms. If left as `None`, the waveform generation
            parameters will be assumed to be the keys from the
            `waveform_sampler.parameter_sampler` object.
    """

    def __init__(
        self,
        dataset_path: Path,
        *args,
        waveform_generation_parameters: Optional[list[str]] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.dataset_path = dataset_path

        # by default, assume waveform generation parameters are those
        # specified in the parameter sampler parameter keys;
        # otherwise, use user passed parameters
        prior_keys = list(
            self.waveform_sampler.parameter_sampler.parameters.keys()
        )

        self.waveform_generation_parameters = (
            prior_keys
            if waveform_generation_parameters is None
            else waveform_generation_parameters
        )

        # only apply parameter sampler conversion function
        # if the waveform_generation_parameters were not specified
        self.convert = waveform_generation_parameters is None
        self.i = 0

    def setup(self, stage: str):
        world_size, rank = self.get_world_size_and_rank()
        self._logger = self.get_logger(world_size, rank)
        if stage != "test":
            raise ValueError(
                "ParameterTestingDataset should only be used for testing"
            )

        parameters = {}
        load = self.hparams.inference_params + ["ra", "dec", "psi", "gpstime"]
        if self.waveform_generation_parameters is not None:
            load += self.waveform_generation_parameters

        with h5py.File(self.dataset_path) as f:
            for parameter in load:
                if parameter == "phi":
                    continue

                parameters[parameter] = torch.tensor(
                    f[parameter][:], dtype=torch.float32
                )

        # apply conversion function to parameters
        # if we're generating from
        if self.convert:
            parameters = (
                self.waveform_sampler.parameter_sampler.conversion_function(
                    parameters
                )
            )

        (
            parameters["mass_1"],
            parameters["mass_2"],
        ) = chirp_mass_and_mass_ratio_to_components(
            parameters["chirp_mass"], parameters["mass_ratio"]
        )
        # convert ra to phi
        parameters["phi"] = torch.tensor(
            phi_from_ra(
                parameters["ra"].numpy(), parameters["gpstime"].numpy()
            )
        )

        # generate cross and plus using our infrastructure
        cross, plus = self.waveform_sampler(**parameters)

        # convert back to tensor
        params = []
        for k in self.hparams.inference_params:
            if k in parameters.keys():
                params.append(torch.Tensor(parameters[k]))

        self.waveforms = torch.stack([cross, plus], dim=0)
        self.parameters = torch.column_stack(params)
        self.background = self.background_from_gpstimes(
            parameters["gpstime"], self.test_fnames
        )

        # once we've generated validation/testing waveforms on cpu,
        # build data augmentation modules
        # and transfer them to appropiate device
        self.build_transforms(stage)
        self.transforms_to_device()

    def test_dataloader(self) -> torch.utils.data.DataLoader:
        cross, plus = self.waveforms

        waveform_dataset = torch.utils.data.TensorDataset(
            cross, plus, self.parameters
        )

        waveform_dataloader = torch.utils.data.DataLoader(
            waveform_dataset,
            batch_size=1,
            shuffle=False,
            pin_memory=False,
            num_workers=10,
        )

        background_dataset = torch.utils.data.TensorDataset(self.background)

        background_dataloader = torch.utils.data.DataLoader(
            background_dataset, pin_memory=False, num_workers=10, shuffle=False
        )
        return ZippedDataset(
            waveform_dataloader,
            background_dataloader,
        )

    def on_after_batch_transfer(self, batch, _):
        """
        Override of the on_after_batch_transfer hook
        defined in `BaseDataset`.
        """

        if not self.trainer.testing:
            raise ValueError(
                "Use of ParameterTestingDataset is for testing only"
            )

        [cross, plus, parameters], [X] = batch

        parameters = {
            k: parameters[:, i]
            for i, k in enumerate(self.hparams.inference_params)
        }

        dec, psi, phi = (
            parameters["dec"].float(),
            parameters["psi"].float(),
            parameters["phi"].float(),
        )
        waveforms = self.projector(dec, psi, phi, cross=cross, plus=plus)

        # downselect to requested inference parameters
        parameters = {
            k: v
            for k, v in parameters.items()
            if k in self.hparams.inference_params
        }

        # make any requested parameter transforms
        parameters = self.transform(parameters)
        parameters = [
            torch.Tensor(parameters[k]) for k in self.hparams.inference_params
        ]
        parameters = torch.vstack(parameters).T

        X, psds = self.psd_estimator(X)
        X += waveforms
        X = self.whitener(X, psds)

        # scale parameters
        parameters = self.scale(parameters)

        # calculate asds and highpass
        freqs = torch.fft.rfftfreq(X.shape[-1], d=1 / self.hparams.sample_rate)
        num_freqs = len(freqs)
        psds = torch.nn.functional.interpolate(
            psds, size=(num_freqs,), mode="linear"
        )

        mask = freqs > self.hparams.highpass
        psds = psds[:, :, mask]
        asds = torch.sqrt(psds)

        return X, asds, parameters


class RawStrainTestingDataset(FlowDataset):
    """
    Testing dataset for analyzing raw strain without injections
    from a given set of gpstimes.

    Useful for testing on real events. This dataset should only be
    used with the `predict` stage.

    Subclass of `amplfi.train.data.datasets.FlowDataset` and
    thus `amplfi.train.data.datasets.BaseDataset`. See those classes
    for additional arguments and keyword arguments.

    Args:
        gpstimes:
            A float, numpy array or path to an hdf5 file containing
            the gps times of the events to analyze. If a path is
            passed, the gps times should be stored in a dataset
            named `gpstimes`
    """

    def __init__(
        self, *args, gpstimes: Union[float, np.ndarray, Path], **kwargs
    ):
        self.gpstimes = self.parse_gps_times(gpstimes)
        super().__init__(*args, **kwargs)

    def parse_gps_times(self, gpstimes: Union[float, np.ndarray, Path]):
        if isinstance(gpstimes, (float, int)):
            gpstimes = np.array([gpstimes])
        elif isinstance(gpstimes, Path):
            with h5py.File(gpstimes, "r") as f:
                gpstimes = f["gpstimes"][:]

        return gpstimes

    def setup(self, stage: str):
        world_size, rank = self.get_world_size_and_rank()
        self._logger = self.get_logger(world_size, rank)
        if stage != "predict":
            raise ValueError(
                "RealEventDataset should only be used with `predict` stage"
            )

        self.background = self.background_from_gpstimes(
            self.gpstimes, self.test_fnames
        )

        # once we've generated validation/testing waveforms on cpu,
        # build data augmentation modules
        # and transfer them to appropiate device
        self.build_transforms(stage)
        self.transforms_to_device()

    def predict_dataloader(self) -> torch.utils.data.DataLoader:
        dataset = torch.utils.data.TensorDataset(
            self.background, self.gpstimes
        )
        dataloader = torch.utils.data.DataLoader(
            dataset,
            pin_memory=False,
            num_workers=10,
            shuffle=False,
            batch_size=1,
        )
        return dataloader

    def on_after_batch_transfer(self, batch, _):
        """
        Override of the on_after_batch_transfer hook
        defined in `BaseDataset`.

        When we're analyzing
        real events, we don't need to do any injections
        """

        [X], gpstimes = batch

        X, psds = self.psd_estimator(X)
        X = self.whitener(X, psds)
        psds = psds[None]

        # calculate asds and highpass
        freqs = torch.fft.rfftfreq(X.shape[-1], d=1 / self.hparams.sample_rate)
        num_freqs = len(freqs)
        psds = torch.nn.functional.interpolate(
            psds, size=(num_freqs,), mode="linear"
        )

        mask = freqs > self.hparams.highpass
        psds = psds[:, :, mask]
        asds = torch.sqrt(psds)

        return X, asds, gpstimes



import torch
import torch.nn as nn

from amplfi.train.architectures.embeddings.base import Embedding


class Expander(nn.Module):
    """Projection head used for VICReg.

    "The projector network has three linear layers, each with 8192 output
    units. The first two layers of the projector are followed by a batch
    normalization layer and rectified linear units." [0]

    - [0]: 2022, VICReg, https://arxiv.org/pdf/2105.04906.pdf
    """

    def __init__(
        self,
        input_dim: int = 2048,
        hidden_dim: int = 8192,
        output_dim: int = 8192,
        num_layers: int = 3,
    ):
        """Initializes the VICRegProjectionHead with the specified dimensions.

        Args:
            input_dim:
                Dimensionality of the input features.
            hidden_dim:
                Dimensionality of the hidden layers.
            output_dim:
                Dimensionality of the output features.
            num_layers:
                Number of layers in the projection head.
        """
        super().__init__()
        blocks = [
            (hidden_dim, hidden_dim, nn.BatchNorm1d(hidden_dim), nn.ReLU())
            for _ in range(num_layers - 2)  # Exclude first and last layer.
        ]
        blocks = [
            (input_dim, hidden_dim, nn.BatchNorm1d(hidden_dim), nn.ReLU()),
            *blocks,
            (hidden_dim, output_dim, None, None),
        ]

        layers = []
        for block in blocks:
            input_dim, output_dim, batch_norm, non_linearity, *bias = block
            use_bias = bias[0] if bias else not bool(batch_norm)
            layers.append(nn.Linear(input_dim, output_dim, bias=use_bias))
            if batch_norm is not None:
                layers.append(batch_norm)
            if non_linearity is not None:
                layers.append(non_linearity)
        self.layers = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Computes one forward pass through the projection head.

        Args:
            x:
                Input of shape bsz x num_ftrs.
        """
        projection: torch.Tensor = self.layers(x)
        return projection


class SimilarityEmbedding(torch.nn.Module):
    """
    Embedding network for similarity learning

    Combines an `embedding` with an expander layer
    as done in https://arxiv.org/pdf/2407.19048
    """

    def __init__(self, embedding: Embedding, expander_factor: int = 4):
        super().__init__()
        self.embedding = embedding
        self.head = Expander(
            input_dim=self.embedding.context_dim,
            hidden_dim=self.embedding.context_dim * expander_factor,
            output_dim=self.embedding.context_dim * expander_factor,
        )

    def forward(self, x):
        x = self.embedding(x)
        x = self.head(x)
        return x



import os
from pathlib import Path

import law
from mldatafind.law.base import DataSandbox

from .paths import paths

root = Path(__file__).resolve().parent.parent.parent
DATA_SANDBOX = f"amplfi::{paths().container_root / 'amplfi.sif'}"

AMPLFI_BIND_DIRS = [
    "AMPLFI_DATADIR",
    "AMPLFI_OUTDIR",
    "AMPLFI_CONDORDIR",
]


class AmplfiDataSandbox(DataSandbox):
    """
    Subclass of `mldatafind.law.base.DataSandbox` for running amplfi tasks;

    Appends amplfi specific environment variables to the container
    and mounts the local amplfi repo into the container in dev mode
    """

    sandbox_type = "amplfi"

    def _get_volumes(self):
        # if running in dev mode, mount the local
        # mldatafind repo into the container so
        # python code changes are reflected
        volumes = super()._get_volumes()
        if self.task and getattr(self.task, "dev", False):
            volumes[str(root)] = "/opt/amplfi"

        # bind AMPLFI env var data directories
        # so users can point to locations
        # that are not automatically mounted;
        # don't bind s3 dirs!

        for path in AMPLFI_BIND_DIRS:
            value = os.getenv(path)
            if value is not None and not value.startswith("s3://"):
                volumes[value] = value
        return volumes

    def _get_env(self):
        env = super()._get_env()

        # append amplfi specific environment variables
        for envvar, value in os.environ.items():
            if envvar.startswith("AMPLFI_"):
                env[envvar] = value

        # forward law config file to the container
        # so tasks can read parameters from it
        env["LAW_CONFIG_FILE"] = os.getenv("LAW_CONFIG_FILE", "")
        return env

    @classmethod
    def config(cls):
        config = super().config()
        config[f"singularity_sandbox_{cls.sandbox_type}"]["forward_law"] = (
            False
        )
        config[f"singularity_sandbox_{cls.sandbox_type}"]["law_executable"] = (
            "/env/bin/law"
        )
        return config


law.config.update(AmplfiDataSandbox.config())


class AmplfiDataTaskMixin:
    """
    Mixin class for appending additional apptainer
    and/or condor environment variable to `mldatafind.law.DataTask`

    This class should be inherited
    alongside a subclass of `mldatafind.law.DataTask`
    """

    def build_environment(self) -> str:
        """
        Append amplfi specific environment variables to the condor submit file
        """
        # get `mldatafind.law.condor.LDGCondorWorkflow` environment variables
        environment = super().build_environment()

        # forward any env variables that start with AMPLFI_
        # that the law config may need to parse
        for envvar, value in os.environ.items():
            if envvar.startswith("AMPLFI_"):
                environment += f"{envvar}={value} "
        environment += '"'
        return environment



from typing import Callable

import torch
from ml4gw.waveforms.generator import TimeDomainCBCWaveformGenerator

from .generator import WaveformGenerator


class CBCGenerator(WaveformGenerator):
    def __init__(
        self,
        *args,
        approximant: Callable,
        f_min: float,
        f_ref: float,
        right_pad: float,
        **kwargs,
    ):
        """
        A lightweight wrapper around
        `ml4gw.waveforms.generator.TimeDomainCBCWaveformGenerator`
        to make it compatible with
        `amplfi.train.data.waveforms.generator.WaveformGenerator`.


        Args:
            *args:
                Positional arguments passed to
                `amplfi.train.data.waveforms.generator.WaveformGenerator`
            approximant:
                A callable that takes parameter tensors
                and returns the cross and plus polarizations.
                For example, `ml4gw.waveforms.IMRPhenomD()`
            f_min:
                Lowest frequency at which waveform signal content
                is generated
            f_ref:
                Reference frequency
            right_pad:
                Position in seconds where coalesence is placed
                relative to the right edge of the window
            **kwargs:
                Keyword arguments passed to
                `amplfi.train.data.waveforms.generator.WaveformGenerator`
        """
        super().__init__(*args, **kwargs)
        self.right_pad = right_pad
        self.approximant = approximant
        self.waveform_generator = TimeDomainCBCWaveformGenerator(
            approximant,
            self.sample_rate,
            self.duration,
            f_min,
            f_ref,
            right_pad + self.fduration / 2,
        )

    def forward(self, **parameters) -> torch.Tensor:
        hc, hp = self.waveform_generator(**parameters)
        waveforms = torch.stack([hc, hp], dim=1)
        if self.time_translator is not None:
            waveforms = self.time_translator(waveforms)
        hc, hp = waveforms.transpose(1, 0)

        return hc.float(), hp.float()



import logging
import os
import time
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from pathlib import Path
from tempfile import gettempdir
from typing import Optional, Tuple, Union

import s3fs
from botocore.exceptions import ClientError, ResponseStreamingError
from filelock import FileLock
from fsspec.exceptions import FSTimeoutError

# s3 retry configuration
retry_config = {"retries": {"total_max_attempts": 10, "mode": "adaptive"}}

# suppress botocore logs
logging.getLogger("botocore").setLevel(logging.WARNING)


def split_data_dir(data_dir: Union[str, Path]) -> Tuple[Optional[str], str]:
    """
    Check if a data directory specifies a remote s3
    source by including  `"s3://"` at the start of
    the directory name.
    """
    data_dir = str(data_dir)
    if data_dir.startswith("s3://"):
        bucket = data_dir.replace("s3://", "")

        # check if specified a target location to map to
        # by adding a colon at the end of our bucket
        bucket, *data_dir = bucket.split(":")
        data_dir = data_dir[0] if data_dir else None
        return bucket, data_dir
    else:
        return None, data_dir


def get_data_dir(data_dir: str) -> Path:
    # generate our local node data directory
    # if our specified data source is remote
    bucket, data_dir = split_data_dir(data_dir)
    if bucket is not None and data_dir is None:
        # we have remote data, but we didn't explicitly
        # specify a directory to download it to, so create
        # a tmp directory using the worker id so that each
        # worker process downloads its own copy of the data
        # only on its first training run
        tmpdir = gettempdir()
        logging.info("Downloading data to local tmp directory")
        data_dir = f"{tmpdir}/data-tmp"

    logging.info(f"Downloading data to {data_dir}")
    os.makedirs(data_dir, exist_ok=True)
    return Path(data_dir)


def _download(
    s3: s3fs.S3FileSystem, source: str, target: str, num_retries: int = 5
):
    """
    Cheap wrapper around s3.get to try to avoid issues
    from interrupted reads.
    """

    lockfile = target + ".lock"
    logging.debug(f"Downloading {source} to {target}")
    for i in range(num_retries):
        with FileLock(lockfile):
            if os.path.exists(target):
                logging.debug(
                    f"Object {source} already downloaded by another process"
                )
                return
            try:
                s3.get(source, target)
                break
            except (ResponseStreamingError, FSTimeoutError, ClientError):
                logging.debug(
                    "Download attempt {} for object {} "
                    "was interrupted, retrying".format(i + 1, source)
                )
                time.sleep(5)
                try:
                    os.remove(target)
                except FileNotFoundError:
                    continue

    else:
        raise RuntimeError(
            "Failed to download object {} due to repeated "
            "connection interruptions".format(source)
        )


def download_training_data(bucket: str, data_dir: str):
    """
    Download s3 data if it doesn't exist.
    """
    data_dir = str(data_dir)
    logging.info(
        "Downloading data from S3 bucket {} to local directory {}".format(
            bucket, data_dir
        )
    )

    # make a local directory to cache data if it
    # doesn't already exist
    background_dir = f"{data_dir}/train/background"
    os.makedirs(background_dir, exist_ok=True)

    # check to make sure the specified bucket
    # actually has data to download
    s3 = s3fs.S3FileSystem(
        key=os.getenv("AWS_ACCESS_KEY_ID"),
        secret=os.getenv("AWS_SECRET_ACCESS_KEY"),
        endpoint_url=os.getenv("AWS_ENDPOINT_URL"),
        config_kwargs=retry_config,
    )
    sources = s3.glob(f"{bucket}/train/background/*.hdf5")
    if not sources:
        raise ValueError(f"No data at {bucket} to download")

    # multiprocess download of training data
    targets = [data_dir + f.replace(f"{bucket}", "") for f in sources]

    download = partial(_download, s3)
    with ThreadPoolExecutor() as executor:
        executor.map(download, sources, targets)



import bilby
from . import skymap
from typing import Optional
from pathlib import Path
import healpy as hp
from astropy import units as u
from astropy.table import Table
from astropy.coordinates import SkyCoord
import matplotlib.pyplot as plt
import numpy as np
from ligo.skymap.postprocess.crossmatch import crossmatch, CrossmatchResult


class AmplfiResult(bilby.result.Result):
    """
    A subclass of `bilby.result.Result` with additional
    convenience methods for generating AMPLFI skymaps
    """

    def to_crossmatch_result(
        self,
        nside: int,
        min_samples_per_pix: int = 15,
        use_distance: bool = True,
        **crossmatch_kwargs,
    ) -> CrossmatchResult:
        """
        Calculate a `ligo.skymap.postprocess.crossmatch.CrossmatchResult`
        based on sky localization and distance posterior samples
        """
        skymap = self.to_skymap(
            nside=nside,
            min_samples_per_pix=min_samples_per_pix,
            use_distance=use_distance,
        )

        coordinates = SkyCoord(
            self.injection_parameters["phi"] * u.rad,
            self.injection_parameters["dec"] * u.rad,
            distance=self.injection_parameters["distance"] * u.Mpc,
        )
        return crossmatch(skymap, coordinates, **crossmatch_kwargs)

    def to_skymap(
        self,
        nside: int,
        min_samples_per_pix: int = 15,
        use_distance: bool = True,
    ) -> Table:
        """Calculate a histogram skymap from posterior samples"""
        distance = None
        if use_distance:
            distance = self.posterior["distance"]

        return skymap.histogram_skymap(
            self.posterior["phi"],
            self.posterior["dec"],
            distance,
            nside=nside,
            min_samples_per_pix=min_samples_per_pix,
        )

    def calculate_searched_area(
        self, nside: int = 32
    ) -> tuple[float, float, float]:
        """
        Calculate the searched area, and estimates
        of 50% and 90% credible region
        """
        skymap = self.to_skymap(nside)["PROBDENSITY"]

        return skymap.calculate_searched_area(
            skymap,
            self.injection_parameters["phi"],
            self.injection_parameters["dec"],
            nside=nside,
        )

    def plot_mollview(
        self, nside: int = 32, outpath: Optional[Path] = None
    ) -> plt.Figure:
        """
        Plot a mollweide projection of the skymap
        """
        skymap = self.to_skymap(nside)["PROBDENSITY"]

        # plot molleweide
        plt.close()
        fig = hp.mollview(skymap, nest=True)

        # plot true values if available
        if self.injection_parameters is not None:
            ra_inj = self.injection_parameters["phi"]
            dec_inj = self.injection_parameters["dec"]
            theta_inj = np.pi / 2 - dec_inj

            hp.visufunc.projscatter(
                theta_inj, ra_inj, marker="x", color="red", s=150
            )

        plt.savefig(outpath)

        return fig



from typing import Tuple

import torch
import torch.nn as nn
from torch import Tensor


class DenseEmbedding(torch.nn.Module):
    """Fully connected embedding with some hidden layers."""

    def __init__(
        self,
        in_features,
        out_features,
        hidden_layer_size=100,
        num_hidden_layers=3,
        activation=torch.nn.functional.relu,
    ) -> None:
        super().__init__()
        self.in_features = in_features
        self.hidden_layer_size = hidden_layer_size
        self.num_hidden_layers = num_hidden_layers
        self.out_features = out_features
        self.activation = activation

        self.initial_layer = nn.Linear(
            self.in_features, self.hidden_layer_size
        )
        self.final_layer = nn.Linear(self.hidden_layer_size, self.out_features)
        self.hidden_layers = nn.ModuleList(
            [
                nn.Linear(self.hidden_layer_size, self.hidden_layer_size)
                for _ in range(self.num_hidden_layers)
            ]
        )

    def forward(self, x):
        x = self.initial_layer(x)
        x = self.activation(x)
        for layer in self.hidden_layers:
            x = self.activation(layer(x))
        x = self.final_layer(x)
        return x


class NChannelDenseEmbedding(nn.Module):
    """
    DenseEmbedding for N channels. Creates a :meth:`DenseEmbedding`
    for individual channel, with a final Linear layer acting on stacked
    embedding on a single channel.
    Expect input shape is (num_batch, N, in_shape). Output shape
    is (num_batch, 1, out_shape)
    """

    def __init__(
        self,
        shape: Tuple[int, int],
        context_dim: int,
    ) -> None:
        super().__init__()
        n_ifos, in_features = shape
        self.activation = torch.nn.functional.relu
        self.embeddings = nn.ModuleList(
            [
                DenseEmbedding(
                    in_features, context_dim, activation=self.activation
                )
                for _ in range(n_ifos)
            ]
        )
        self.final_layer = DenseEmbedding(
            n_ifos * context_dim, context_dim, activation=self.activation
        )

    def forward(self, x):
        embedded_vals = []
        for channel_num, embedding in enumerate(self.embeddings):
            embedded_vals.append(embedding(x[:, channel_num, :]))

        x_concat = torch.concat(embedded_vals, dim=1)
        x_concat = self.final_layer(x_concat)
        x_concat = self.activation(x_concat)
        return x_concat


class BasicBlock(torch.nn.Module):
    def __init__(
        self,
        size: int,
        n_channels: int,
    ):
        super().__init__()
        self.activation = torch.nn.functional.relu
        self.layer = nn.Linear(size, size)
        self.norm = nn.BatchNorm1d(n_channels)

    def forward(self, x: Tensor) -> Tensor:
        out = self.layer(x)
        out = self.norm(out)
        out = self.activation(out)

        return out


class CoherentDenseEmbedding(torch.nn.Module):
    """Fully connected embedding
    Embeds a Tensor of shape (num_batch, n_ifos, n_samples)
    into a Tensor of shape (num_batch, 1, out_features)
    through a fully connected network with num_hidden_layers
    hidden layers of size hidden_layer_size.
    After passing through the hidden layers, the tensor is passed
    through a penultimate layer that will output a Tensor of shape
    (num_batch, n_ifos, out_features). Finally, the tensor is concated along
    the ifo dimension and passed through a final layer that
    will output a Tensor of shape (num_batch, 1, out_features).
    Args:
        shape: A (n_ifos, in_features) Tuple
        context_dim: Number of total output features
        hidden_layer_size: Number of hidden units in each hidden layer
        num_hidden_layers: Number of hidden layers
        activation: Activation function to use between layers
    """

    def __init__(
        self,
        shape: Tuple[int, int],
        context_dim: int,
        hidden_layer_size: int,
        num_hidden_layers: int,
    ) -> None:
        super().__init__()
        n_ifos, in_features = shape
        self.activation = torch.nn.functional.relu

        self.initial_layer = nn.Linear(in_features, hidden_layer_size)
        self.hidden_layers = nn.Sequential(
            *[
                BasicBlock(
                    hidden_layer_size, n_ifos, activation=self.activation
                )
                for _ in range(num_hidden_layers)
            ]
        )
        self.fc1 = nn.Linear(hidden_layer_size, context_dim)
        self.fc2 = nn.Linear(n_ifos * context_dim, context_dim)

    def forward(self, x):
        x = self.initial_layer(x)
        x = self.activation(x)
        x = self.hidden_layers(x)
        x = self.fc1(x)
        x = x.reshape(len(x), 1, -1)
        x = self.activation(x)
        x = self.fc2(x)
        x = x.reshape(len(x), -1)
        return x



import law
import luigi
from mldatafind.law.parameters import PathParameter
from mldatafind.law.tasks import Fetch
from mldatafind.law.tasks import Query as _Query
from mldatafind.law.tasks.condor.workflows import StaticMemoryWorkflow
from .base import DATA_SANDBOX, AmplfiDataTaskMixin
from .paths import paths


# add mixin for appending amplfi specific
# environment variables to condor and containers
class Query(AmplfiDataTaskMixin, _Query):
    pass


# override FetchTrain and FetchTest
# tasks to use Query with mixin
class FetchTrain(AmplfiDataTaskMixin, Fetch):
    def workflow_requires(self):
        reqs = {}
        reqs["segments"] = Query.req(self, segments_file=self.segments_file)
        return reqs


class FetchTest(AmplfiDataTaskMixin, Fetch):
    def workflow_requires(self):
        reqs = {}
        reqs["segments"] = Query.req(self, segments_file=self.segments_file)
        return reqs


class DataGeneration(law.WrapperTask):
    """
    Pipeline for launching FetchTrain and FetchTest tasks
    """

    dev = luigi.BoolParameter(
        default=False, description="Run the task in development mode."
    )

    def requires(self):
        yield FetchTrain.req(
            self,
            sandbox=DATA_SANDBOX,
            data_dir=paths().data_dir / "train" / "background",
            condor_directory=paths().condor_dir / "train",
        )

        yield FetchTest.req(
            self,
            sandbox=DATA_SANDBOX,
            data_dir=paths().data_dir / "test" / "background",
            condor_directory=paths().condor_dir / "test",
        )


class LigoSkymap(
    AmplfiDataTaskMixin,
    law.LocalWorkflow,
    StaticMemoryWorkflow,
    law.SandboxTask,
):
    """
    Workflow for parallelizing skymap generation via ligo-skymap-from-samples
    """

    data_dir = PathParameter(
        description="Path to the directory containing the "
        "event sub directories."
        "Each sub directory should contain "
        "a posterior_samples.dat file."
    )
    ligo_skymap_args = luigi.OptionalListParameter(
        description="Additional command line style arguments"
        "to pass to ligo-skymap-from-samples.",
        default="",
    )
    dev = luigi.BoolParameter(
        default=False, description="Run the task in development mode."
    )
    sandbox = DATA_SANDBOX

    def sandbox_env(self, env):
        env = super().sandbox_env(env)
        env.update({"MKL_NUM_THREADS": "1", "OMP_NUM_THREADS": "1"})
        return env

    def create_branch_map(self):
        branch_map, i = {}, 0
        for event_dir in self.data_dir.iterdir():
            if event_dir.is_dir():
                branch_map[i] = event_dir / "posterior_samples.dat"
                i += 1
        return branch_map

    def output(self):
        event_dir = self.branch_data.parent
        return law.LocalFileTarget(event_dir / "skymap.fits")

    def run(self):
        from ligo.skymap.tool import (
            ligo_skymap_from_samples,
        )

        args = [
            str(self.branch_data),
            "-j",
            str(self.request_cpus),
            "-o",
            str(self.branch_data.parent),
        ]
        if self.ligo_skymap_args:
            args.extend(self.ligo_skymap_args)

        # call ligo-skymap-from-samples
        ligo_skymap_from_samples.main(args)



from pathlib import Path
from typing import TYPE_CHECKING
from gwpy.plot import Plot
from gwpy.timeseries import TimeSeries
import lightning.pytorch as pl
import matplotlib.pyplot as plt
import numpy as np

import bilby
from tqdm.auto import tqdm


if TYPE_CHECKING:
    from ligo.skymap.postprocess.crossmatch import CrossmatchResult
    from amplfi.train.models import FlowModel


class StrainVisualization(pl.Callback):
    """
    Lightning Callback for visualizing the strain data
    and asds being analyzed during the test step
    """

    def __init__(self, outdir: Path, num_plot: int):
        self.outdir = outdir
        self.num_plot = num_plot

    def on_test_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx=0
    ) -> None:
        """
        Called at the end of each test step.
        `outputs` consists of objects returned by `pl_module.test_step`.
        """

        # test_step returns bilby result object
        result = outputs

        if batch_idx >= self.num_plot:
            return

        outdir = self.outdir / f"event_{batch_idx}"
        outdir.mkdir(exist_ok=True)

        # unpack batch
        strain, asds, _ = batch
        strain, asds = strain[0].cpu().numpy(), asds[0].cpu().numpy()

        # steal some attributes needed from datamodule
        # TODO: should we link these to the pl_module?
        highpass = trainer.datamodule.hparams.highpass
        sample_rate = trainer.datamodule.hparams.sample_rate
        ifos = trainer.datamodule.hparams.ifos

        # filenames for various plots
        whitened_td_strain_fname = outdir / "whitened_td.png"
        whitened_fd_strain_fname = outdir / "whitened_fd.png"
        qscan_fname = outdir / "qscan.png"
        asd_fname = outdir / "asd.png"

        # whitened time domain strain
        plt.figure()
        plt.title("Whitened Time Domain Strain")

        for i, ifo in enumerate(ifos):
            plt.plot(strain[i], label=ifo)

        plt.legend()
        plt.savefig(whitened_td_strain_fname)
        plt.close()

        # qscans
        qscans = []

        for i in range(len(ifos)):
            ts = TimeSeries(
                strain[i],
                dt=1 / sample_rate,
            )

            spec = ts.q_transform(
                logf=True,
                whiten=False,
                frange=(25, 500),
                qrange=(4, 108),
                # outseg=(3, 3.7),
            )
            qscans.append(spec)

        title = ""
        if result.injection_parameters is not None:
            chirp_mass, mass_ratio = (
                result.injection_parameters["chirp_mass"],
                result.injection_parameters["mass_ratio"],
            )
            title = f"chirp_mass: {chirp_mass:2f}, mass_ratio: {mass_ratio:2f}"
        plot = Plot(
            *qscans,
            figsize=(18, 5),
            geometry=(1, 2),
            yscale="log",
            method="pcolormesh",
            cmap="viridis",
            title=title,
        )

        for i, ax in enumerate(plot.axes):
            label = "" if i != 1 else "Normalized Energy"
            plot.colorbar(ax=ax, label=label)

        plot.savefig(qscan_fname)
        plt.close()

        # asds
        frequencies = trainer.datamodule.frequencies.numpy()
        mask = frequencies > highpass
        frequencies_masked = frequencies[mask]
        plt.figure()
        for i, ifo in enumerate(ifos):
            plt.loglog(frequencies_masked, asds[i], label=f"{ifo} asd")
            plt.title("ASDs")
            plt.xlabel("Frequency (Hz)")
            plt.ylabel("Scaled Amplitude")

        plt.legend()
        plt.savefig(asd_fname)
        plt.close()

        # whitened frequency domain data
        fig, axes = plt.subplots(1, 2, figsize=(14, 6))
        for i, ifo in enumerate(ifos):
            strain_fft = np.fft.rfft(strain[i])
            freqs = np.fft.rfftfreq(n=strain[i].shape[-1], d=1 / sample_rate)

            axes[0].plot(
                freqs,
                strain_fft.real / (sample_rate ** (1 / 2)),
                label=f"{ifo} data",
            )
            axes[1].plot(
                freqs,
                strain_fft.imag / (sample_rate ** (1 / 2)),
                label=f"{ifo} data",
            )

        axes[0].set_title("Real")
        axes[1].set_title("Imaginary")

        axes[0].set_ylabel("Whitened Amplitude")
        axes[0].set_xlabel("Frequency (Hz)")
        axes[0].set_xlabel("Frequency (Hz)")

        plt.legend()

        plt.savefig(whitened_fd_strain_fname)
        plt.close()

    def on_predict_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx=0
    ):
        return self.on_test_batch_end(
            trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
        )


class SavePosterior(pl.Callback):
    """
    Lightning Callback for bilby result objects and
    posterior data to disk
    """

    def __init__(self, outdir: Path):
        self.outdir = outdir

    def on_test_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx=0
    ) -> None:
        """
        Called at the end of each test step.
        `outputs` consists of objects returned by `pl_module.test_step`.
        """

        # test_step returns bilby result object
        result = outputs

        outdir = self.outdir / f"event_{batch_idx}"
        outdir.mkdir(exist_ok=True)

        # save posterior samples for ease of use with
        # ligo skymap and save full result to have
        # access to the true injection parameters
        result.save_posterior_samples(outdir / "posterior_samples.dat")
        result.save_to_file(outdir / "result.hdf5", extension="hdf5")

    def on_predict_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx=0
    ):
        return self.on_test_batch_end(
            trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
        )


class ProbProbPlot(pl.Callback):
    def on_test_epoch_end(self, _, pl_module: "FlowModel"):
        logger = pl_module._logger
        logger.info("Making PP plot")
        bilby.result.make_pp_plot(
            pl_module.test_results,
            save=True,
            filename=pl_module.test_outdir / "pp-plot.png",
            keys=pl_module.inference_params,
        )


class CrossMatchStatistics(pl.Callback):
    """
    Use `ligo.skymap.postprocess.crossmatch` to calculate skymap statistics.
    Make searched area and searched volume CDF plots
    """

    def on_test_epoch_end(self, _, pl_module: "FlowModel"):
        crossmatch_results: list["CrossmatchResult"] = []
        test_outdir = pl_module.test_outdir
        logger = pl_module._logger

        logger.info("Calculating cross match statistics for each result")
        for result in tqdm(pl_module.test_results):
            # calculate skymap staistics via ligo.skymap.postprocess.crossmatch
            crossmatch_result = result.to_crossmatch_result(
                nside=pl_module.nside,
                min_samples_per_pix=pl_module.min_samples_per_pix,
                use_distance=True,
                contours=[0.5, 0.9],
            )
            crossmatch_results.append(crossmatch_result)

        # searched area cum hist
        searched_areas = [
            result.searched_area for result in crossmatch_results
        ]
        searched_volumes = [
            result.searched_vol for result in crossmatch_results
        ]
        fifty_percent_areas = [
            result.contour_areas[0] for result in crossmatch_results
        ]
        ninety_percent_areas = [
            result.contour_areas[1] for result in crossmatch_results
        ]

        searched_areas = np.sort(searched_areas)
        searched_volumes = np.sort(searched_volumes)
        counts = np.arange(1, len(searched_areas) + 1) / len(searched_areas)

        plt.figure(figsize=(10, 6))
        plt.step(searched_volumes, counts, where="post")
        plt.xscale("log")
        plt.xlabel("Searched Volume (Mpc^3)")
        plt.ylabel("Cumulative Probability")
        plt.title("Searched Volume Cumulative Distribution Function")
        plt.grid()
        plt.axhline(0.5, color="grey", linestyle="--")
        plt.savefig(test_outdir / "searched_volume.png")
        np.save(test_outdir / "searched_volume.npy", searched_volumes)

        plt.figure(figsize=(10, 6))
        plt.step(searched_areas, counts, where="post")
        plt.xscale("log")
        plt.xlabel("Searched Area (deg^2)")
        plt.ylabel("Cumulative Probability")
        plt.title("Searched Area Cumulative Distribution Function")
        plt.grid()
        plt.axhline(0.5, color="grey", linestyle="--")
        plt.savefig(test_outdir / "searched_area.png")
        np.save(test_outdir / "searched_area.npy", searched_areas)

        plt.close()
        plt.figure(figsize=(10, 6))
        mm, bb, pp = plt.hist(
            fifty_percent_areas, label="50 percent area", bins=50
        )
        _, _, _ = plt.hist(
            ninety_percent_areas, label="90 percent area", bins=bb
        )
        plt.xlabel("Sq. deg.")
        plt.legend()
        plt.savefig(test_outdir / "fifty_ninety_areas.png")



import logging
import math
import sys
from pathlib import Path
from typing import Optional

import lightning.pytorch as pl
import torch
from ml4gw.transforms import ChannelWiseScaler

Tensor = torch.Tensor
Distribution = torch.distributions.Distribution


class AmplfiModel(pl.LightningModule):
    """
    Amplfi model base class

    Encodes common functionality for all models,
    such as on-device augmentation and preprocessing,

    Args:
        checkpoint:
            Path to a model checkpoint to load. This will load in weights
            for both flow and embedding. Should only be specified when
            running `trainer.test`. For resuming a `trainer.fit` run from
            a checkpoint, use the --ckpt_path `Trainer` argument.
    """

    def __init__(
        self,
        inference_params: list[str],
        learning_rate: float,
        train_outdir: Path,
        test_outdir: Optional[Path] = None,
        weight_decay: float = 0.0,
        patience: int = 10,
        factor: float = 0.1,
        verbose: bool = False,
    ):
        super().__init__()
        self.scheduler_patience = patience
        self.scheduler_factor = factor
        self._logger = self.init_logging(verbose)

        if test_outdir is None:
            test_outdir = train_outdir / "test_results"

        train_outdir.mkdir(exist_ok=True)
        test_outdir.mkdir(exist_ok=True)
        self.test_outdir = test_outdir
        self.train_outdir = train_outdir

        self.inference_params = inference_params

        self.save_hyperparameters()

        # initialize an unfit scaler here so that it is available
        # for the LightningModule to save and load from checkpoints
        self.scaler = ChannelWiseScaler(len(inference_params))

    def init_logging(self, verbose):
        log_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        logging.basicConfig(
            format=log_format,
            level=logging.DEBUG if verbose else logging.INFO,
            stream=sys.stdout,
        )

        world_size, rank = self.get_world_size_and_rank()
        logger_name = self.__class__.__name__
        if world_size > 1:
            logger_name += f":{rank}"
        return logging.getLogger(logger_name)

    def get_world_size_and_rank(self) -> tuple[int, int]:
        """
        Name says it all, but generalizes to the case
        where we aren't running distributed training.
        """
        if not torch.distributed.is_initialized():
            return 1, 0
        else:
            world_size = torch.distributed.get_world_size()
            rank = torch.distributed.get_rank()
            return world_size, rank

    def setup(self, stage):
        if stage == "fit":
            # if we're fitting, store an instance of the
            # fit scaler in the model
            # so that its weights can be checkpointed
            self.scaler = self.trainer.datamodule.scaler

    def configure_optimizers(self):
        if not torch.distributed.is_initialized():
            world_size = 1
        else:
            world_size = torch.distributed.get_world_size()

        lr = self.hparams.learning_rate * math.sqrt(world_size)
        optimizer = torch.optim.AdamW(
            self.model.parameters(),
            lr=lr,
            weight_decay=self.hparams.weight_decay,
        )

        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            optimizer,
            patience=self.scheduler_patience,
            factor=self.scheduler_factor,
        )

        return {
            "optimizer": optimizer,
            "lr_scheduler": {"scheduler": scheduler, "monitor": "valid_loss"},
        }

    def scale(self, parameters, reverse: bool = False):
        """
        Apply standard scaling to transformed parameters
        """
        parameters = parameters.transpose(1, 0)
        scaled = self.scaler(parameters, reverse=reverse)
        scaled = scaled.transpose(1, 0)
        return scaled



from typing import Optional, Tuple

import torch
from ml4gw import gw
from ml4gw.transforms import SpectralDensity

Tensor = torch.Tensor


class WaveformProjector(torch.nn.Module):
    def __init__(
        self,
        ifos: list[str],
        sample_rate: float,
    ) -> None:
        super().__init__()
        tensors, vertices = gw.get_ifo_geometry(*ifos)
        self.register_buffer("tensors", tensors)
        self.register_buffer("vertices", vertices)

        self.sample_rate = sample_rate

    def forward(
        self,
        dec: torch.Tensor,
        psi: torch.Tensor,
        phi: torch.Tensor,
        **polarizations: torch.Tensor,
    ) -> torch.Tensor:
        responses = gw.compute_observed_strain(
            dec,
            psi,
            phi,
            detector_tensors=self.tensors,
            detector_vertices=self.vertices,
            sample_rate=self.sample_rate,
            **polarizations,
        )
        return responses


class PsdEstimator(torch.nn.Module):
    """
    Module that takes a sample of data, splits it into
    two unequal-length segments, calculates the PSD of
    the first section, then returns this PSD along with
    the second section.

    Args:
        length:
            The length, in seconds, of timeseries data
            to be returned for whitening. Note that the
            length of time used for the PSD will then be
            whatever remains along first part of the time
            axis of the input.
        sample_rate:
            Rate at which input data has been sampled in Hz
        fftlength:
            Length of FFTs to use when computing the PSD
        overlap:
            Amount of overlap between FFT windows when
            computing the PSD. Default value of `None`
            uses `fftlength / 2`
        average:
            Method for aggregating spectra from FFT
            windows, either `"mean"` or `"median"`
        fast:
            If `True`, use a slightly faster PSD algorithm
            that is inaccurate for the lowest two frequency
            bins. If you plan on highpassing later, this
            should be fine.
    """

    def __init__(
        self,
        length: float,
        sample_rate: float,
        fftlength: float,
        overlap: Optional[float] = None,
        average: str = "mean",
        fast: bool = True,
    ) -> None:
        super().__init__()
        self.size = int(length * sample_rate)
        self.spectral_density = SpectralDensity(
            sample_rate, fftlength, overlap, average, fast=fast
        )

    def forward(self, X: Tensor) -> Tuple[Tensor, Tensor]:
        splits = [X.size(-1) - self.size, self.size]
        background, X = torch.split(X, splits, dim=-1)

        # if we have 2 batch elements in our input data,
        # it will be assumed that the 0th element corresponds
        # to true background and the 1th element corresponds
        # to injected data, in which case we'll only compute
        # the background PSD on the former
        if X.ndim == 3 and X.size(0) == 2:
            background = background[0]

        self.spectral_density.to(device=background.device)
        psds = self.spectral_density(background.double())
        return X, psds



from ..data.datasets.base import AmplfiDataset
from ..models.base import AmplfiModel
from .base import AmplfiBaseCLI


class AmplfiSimilarityCLI(AmplfiBaseCLI):
    def add_arguments_to_parser(self, parser):
        super().add_arguments_to_parser(parser)

        parser.link_arguments(
            "data.init_args.ifos",
            "model.init_args.arch.init_args.embedding.init_args.num_ifos",
            compute_fn=lambda x: len(x),
            apply_on="parse",
        )
        return parser


def main(args=None):
    cli = AmplfiSimilarityCLI(
        AmplfiModel,
        AmplfiDataset,
        subclass_mode_model=True,
        subclass_mode_data=True,
        save_config_kwargs={"overwrite": True},
        seed_everything_default=101588,
        args=args,
    )
    return cli


if __name__ == "__main__":
    main()



from typing import Optional

import torch

from ..augmentors import TimeTranslator
from ..utils.utils import ParameterTransformer

Distribution = torch.distributions.Distribution


class WaveformSampler(torch.nn.Module):
    """
    Base object defining methods that waveform producing classes
    should implement. Should not be instantiated on its own.

    Args:
        fduration:
            Desired length in seconds of the time domain
            response of the whitening filter built from PSDs.
            See `ml4gw.spectral.truncate_inverse_power_spectrum`
        kernel_length:
            Length in seconds of window passed to neural network.
        sample_rate:
            Sample rate in Hz of generated waveforms
        inference_params:
            The parameters the model will perform inference on
        dec:
            The distribution of declinations to sample from
        psi:
            The distribution of polarization angles to sample from
        phi:
            The distribution of "right ascensions" to sample from
        jitter:
            The amount of jitter in seconds to randomly shift
            the waveform coalescence time. If `None`, no jitter is applied.
        parameter_transformer:
            A `ParameterTransformer` object that applies any
            additional transformations to parameters before
            they are scaled and passed to the neural network.

    """

    def __init__(
        self,
        *args,
        fduration: float,
        kernel_length: float,
        sample_rate: float,
        inference_params: list[str],
        dec: Distribution,
        psi: Distribution,
        phi: Distribution,
        jitter: Optional[float] = None,
        parameter_transformer: Optional[ParameterTransformer] = None,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.parameter_transformer = parameter_transformer or (lambda x: x)
        self.inference_params = inference_params
        self.fduration = fduration
        self.kernel_length = kernel_length

        self.sample_rate = sample_rate
        self.dec, self.psi, self.phi = dec, psi, phi
        self.time_translator = (
            TimeTranslator(jitter, sample_rate) if jitter is not None else None
        )

    def sample_extrinsic(self, X: torch.Tensor):
        """
        Sample extrinsic parameters used to project waveforms
        """
        N = len(X)
        dec = self.dec.sample((N,)).to(X.device)
        psi = self.psi.sample((N,)).to(X.device)
        phi = self.phi.sample((N,)).to(X.device)
        return dec, psi, phi

    @property
    def duration(self):
        """
        Length of kernel before whitening removes
        fduration / 2 from each side
        """
        return self.fduration + self.kernel_length

    def get_val_waveforms(self):
        raise NotImplementedError

    def get_test_waveforms(self):
        raise NotImplementedError

    def fit_scaler(self, scaler: torch.nn.Module) -> torch.nn.Module:
        raise NotImplementedError

    def sample(self, X: torch.Tensor):
        """Defines how to sample waveforms for training"""
        raise NotImplementedError



from typing import Callable, Dict, Optional

import torch


class ParameterTransformer(torch.nn.Module):
    """
    Helper class for applying preprocessing
    transformations to inference parameters
    """

    def __init__(self, **transforms: Callable):
        super().__init__()
        self.transforms = transforms

    def forward(
        self,
        parameters: Dict[str, torch.Tensor],
    ):
        # transform parameters
        transformed = {k: v(parameters[k]) for k, v in self.transforms.items()}
        # update parameter dict
        parameters.update(transformed)
        return parameters


class ParameterSampler(torch.nn.Module):
    def __init__(
        self,
        conversion_function: Optional[Callable] = None,
        **parameters: Callable,
    ):
        """
        A class for sampling parameters from a prior distribution

        Args:
            conversion_function:
                A callable that takes a dictionary of sampled parameters
                and returns a dictionary of waveform generation parameters
            **parameters:
                A dictionary of parameter samplers that take an integer N
                and return a tensor of shape (N, ...) representing
                samples from the prior distribution
        """
        super().__init__()
        self.parameters = parameters
        self.conversion_function = conversion_function or (lambda x: x)

    def forward(
        self,
        N: int,
        device: str = "cpu",
    ):
        # sample parameters from prior
        parameters = {
            k: v.sample((N,)).to(device) for k, v in self.parameters.items()
        }
        # perform any necessary conversions
        # to from sampled parameters to
        # waveform generation parameters
        parameters = self.conversion_function(parameters)
        return parameters


class ZippedDataset(torch.utils.data.IterableDataset):
    def __init__(self, *datasets):
        super().__init__()
        self.datasets = datasets

    def __len__(self):
        lengths = []
        for dset in self.datasets:
            try:
                lengths.append(len(dset))
            except Exception as e:
                raise e from None
        return min(lengths)

    def __iter__(self):
        return zip(*self.datasets, strict=False)



import torch

from .base import AmplfiDataset


class FlowDataset(AmplfiDataset):
    """
    Lightning DataModule for training normalizing flow networks
    """

    def inject(self, X, cross, plus, parameters):
        self.projector.to(self.device)
        self.whitener.to(self.device)

        X, psds = self.psd_estimator(X)
        dec, psi, phi = self.waveform_sampler.sample_extrinsic(X)

        waveforms = self.projector(dec, psi, phi, cross=cross, plus=plus)

        # append extrinsic parameters to parameters
        parameters.update({"dec": dec, "psi": psi, "phi": phi})

        # downselect to requested inference parameters
        parameters = {
            k: v
            for k, v in parameters.items()
            if k in self.hparams.inference_params
        }

        # make any requested parameter transforms
        parameters = self.transform(parameters)
        parameters = [
            torch.Tensor(parameters[k]) for k in self.hparams.inference_params
        ]
        parameters = torch.vstack(parameters).T
        X += waveforms
        X = self.whitener(X, psds)

        # scale parameters
        parameters = self.scale(parameters)

        # calculate asds and highpass
        freqs = torch.fft.rfftfreq(X.shape[-1], d=1 / self.hparams.sample_rate)
        num_freqs = len(freqs)
        psds = torch.nn.functional.interpolate(
            psds, size=(num_freqs,), mode="linear"
        )

        mask = freqs > self.hparams.highpass
        psds = psds[:, :, mask]
        asds = torch.sqrt(psds)

        return X, asds, parameters



from math import pi

import torch
from ml4gw import distributions
from ml4gw.waveforms.conversion import (
    bilby_spins_to_lalsim,
    chirp_mass_and_mass_ratio_to_components,
)
from torch.distributions import Uniform

from .data.utils.utils import ParameterSampler, ParameterTransformer

sg_transformer = ParameterTransformer(hrss=torch.log)

# make the below callables that return parameter samplers
# so that jsonargparse can serialize them properly


# prior and parameter transformer for sg use case
def sg_prior() -> ParameterSampler:
    return ParameterSampler(
        frequency=Uniform(32, 1024),
        quality=Uniform(2, 100),
        hrss=distributions.LogUniform(1e-23, 5e-20),
        phase=Uniform(0, 2 * pi),
        eccentricity=Uniform(0, 1),
    )


def precessing_to_lalsimulation_parameters(
    parameters: dict[str, torch.Tensor],
) -> dict[str, torch.Tensor]:
    """
    Convert precessing spin parameters to lalsimulation parameters
    """
    chirp_mass, mass_ratio = (
        parameters.pop("chirp_mass"),
        parameters.pop("mass_ratio"),
    )
    mass_1, mass_2 = chirp_mass_and_mass_ratio_to_components(
        chirp_mass, mass_ratio
    )

    a_1, a_2 = parameters.pop("a_1"), parameters.pop("a_2")
    tilt_1, tilt_2 = parameters.pop("tilt_1"), parameters.pop("tilt_2")
    inclination = parameters.pop("inclination")
    phi_jl, phi_12 = parameters.pop("phi_jl"), parameters.pop("phi_12")
    phic = parameters.pop("phic")

    # TODO: hard coding f_ref = 40 here b/c not sure best way to link this
    # to the f_ref specified in the config file
    incl, s1x, s1y, s1z, s2x, s2y, s2z = bilby_spins_to_lalsim(
        inclination,
        phi_jl,
        tilt_1,
        tilt_2,
        phi_12,
        a_1,
        a_2,
        mass_1,
        mass_2,
        40,
        torch.zeros(len(inclination), device=inclination.device),
    )

    output = {}
    output["mass_1"] = mass_1
    output["mass_2"] = mass_2
    output["chirp_mass"] = chirp_mass
    output["mass_ratio"] = mass_ratio
    output["s1x"] = s1x
    output["s1y"] = s1y
    output["s1z"] = s1z
    output["s2x"] = s2x
    output["s2y"] = s2y
    output["s2z"] = s2z
    output["distance"] = parameters["distance"]
    output["inclination"] = incl
    output["phic"] = phic
    return output


def aligned_to_lalsimulation_parameters(
    parameters: dict[str, torch.Tensor],
) -> dict[str, torch.Tensor]:
    """
    Convert precessing spin parameters to lalsimulation parameters
    """
    chirp_mass, mass_ratio = (
        parameters.pop("chirp_mass"),
        parameters.pop("mass_ratio"),
    )
    mass_1, mass_2 = chirp_mass_and_mass_ratio_to_components(
        chirp_mass, mass_ratio
    )

    s1x = torch.zeros_like(mass_1)
    s1y = torch.zeros_like(mass_1)

    s2x = torch.zeros_like(mass_1)
    s2y = torch.zeros_like(mass_1)

    output = {}
    output["chi1"] = parameters["chi1"]
    output["chi2"] = parameters["chi2"]
    output["mass_1"] = mass_1
    output["mass_2"] = mass_2
    output["chirp_mass"] = chirp_mass
    output["mass_ratio"] = mass_ratio
    output["s1x"] = s1x
    output["s1y"] = s1y
    output["s1z"] = parameters.pop("chi1")
    output["s2x"] = s2x
    output["s2y"] = s2y
    output["s2z"] = parameters.pop("chi2")
    output["distance"] = parameters.pop("distance")
    output["inclination"] = parameters.pop("inclination")
    output["phic"] = parameters.pop("phic")
    return output


# TODO: we want validate_args = False at test time
# but True at train time


# priors and parameter transformers for cbc use case
def aligned_cbc_prior() -> ParameterSampler:
    """
    Prior for aligned-spin CBC waveform generation, e.g with IMRPhenomD
    """
    return ParameterSampler(
        conversion_function=aligned_to_lalsimulation_parameters,
        chirp_mass=Uniform(
            torch.as_tensor(10, dtype=torch.float32),
            torch.as_tensor(100, dtype=torch.float32),
            validate_args=False,
        ),
        mass_ratio=Uniform(
            torch.as_tensor(0.125, dtype=torch.float32),
            torch.as_tensor(0.999, dtype=torch.float32),
            validate_args=False,
        ),
        distance=Uniform(
            torch.as_tensor(100, dtype=torch.float32),
            torch.as_tensor(3100, dtype=torch.float32),
            validate_args=False,
        ),
        inclination=distributions.Sine(
            torch.as_tensor(0, dtype=torch.float32),
            torch.as_tensor(torch.pi, dtype=torch.float32),
            validate_args=False,
        ),
        phic=Uniform(
            torch.as_tensor(0, dtype=torch.float32),
            torch.as_tensor(2 * torch.pi, dtype=torch.float32),
            validate_args=False,
        ),
        chi1=Uniform(
            torch.as_tensor(-0.999, dtype=torch.float32),
            torch.as_tensor(0.999, dtype=torch.float32),
            validate_args=False,
        ),
        chi2=Uniform(
            torch.as_tensor(-0.999, dtype=torch.float32),
            torch.as_tensor(0.999, dtype=torch.float32),
            validate_args=False,
        ),
    )


def precessing_cbc_prior() -> ParameterSampler:
    """
    Prior for precessing-spin CBC waveform generation, e.g with IMRPhenomPv2
    """

    return ParameterSampler(
        conversion_function=precessing_to_lalsimulation_parameters,
        chirp_mass=Uniform(
            torch.as_tensor(10, dtype=torch.float32),
            torch.as_tensor(100, dtype=torch.float32),
            validate_args=False,
        ),
        mass_ratio=Uniform(
            torch.as_tensor(0.125, dtype=torch.float32),
            torch.as_tensor(0.999, dtype=torch.float32),
            validate_args=False,
        ),
        distance=Uniform(
            torch.as_tensor(100, dtype=torch.float32),
            torch.as_tensor(3100, dtype=torch.float32),
            validate_args=False,
        ),
        inclination=distributions.Sine(
            torch.as_tensor(0, dtype=torch.float32),
            torch.as_tensor(torch.pi, dtype=torch.float32),
            validate_args=False,
        ),
        phic=Uniform(
            torch.as_tensor(0, dtype=torch.float32),
            torch.as_tensor(2 * torch.pi, dtype=torch.float32),
            validate_args=False,
        ),
        a_1=Uniform(
            torch.as_tensor(0, dtype=torch.float32),
            torch.as_tensor(0.999, dtype=torch.float32),
            validate_args=False,
        ),
        a_2=Uniform(
            torch.as_tensor(0, dtype=torch.float32),
            torch.as_tensor(0.999, dtype=torch.float32),
            validate_args=False,
        ),
        tilt_1=Uniform(
            torch.as_tensor(0, dtype=torch.float32),
            torch.as_tensor(pi, dtype=torch.float32),
            validate_args=False,
        ),
        tilt_2=Uniform(
            torch.as_tensor(0, dtype=torch.float32),
            torch.as_tensor(pi, dtype=torch.float32),
            validate_args=False,
        ),
        phi_jl=Uniform(
            torch.as_tensor(0, dtype=torch.float32),
            torch.as_tensor(2 * torch.pi, dtype=torch.float32),
            validate_args=False,
        ),
        phi_12=Uniform(
            torch.as_tensor(0, dtype=torch.float32),
            torch.as_tensor(2 * torch.pi, dtype=torch.float32),
            validate_args=False,
        ),
    )


def cbc_testing_delta_function_prior() -> ParameterSampler:
    return ParameterSampler(
        chirp_mass=distributions.DeltaFunction(
            torch.as_tensor(55, dtype=torch.float32),
            validate_args=False,
        ),
        mass_ratio=distributions.DeltaFunction(
            torch.as_tensor(0.9, dtype=torch.float32),
            validate_args=False,
        ),
        distance=distributions.DeltaFunction(
            torch.as_tensor(1000, dtype=torch.float32),
            validate_args=False,
        ),
        inclination=distributions.DeltaFunction(
            torch.as_tensor(torch.pi / 6, dtype=torch.float32),
            validate_args=False,
        ),
        phic=distributions.DeltaFunction(
            torch.as_tensor(torch.pi, dtype=torch.float32),
            validate_args=False,
        ),
        chi1=distributions.DeltaFunction(
            torch.as_tensor(0, dtype=torch.float32),
            validate_args=False,
        ),
        chi2=distributions.DeltaFunction(
            torch.as_tensor(0, dtype=torch.float32),
            validate_args=False,
        ),
    )



import torch
import torch.nn.functional as F


class TimeTranslator(torch.nn.Module):
    """
    A torch.nn.Module that shifts waveforms in time

    Args:
        jitter:
            The maximum magnitude of time to shift waveforms in seconds.
            Waveforms will be shifted by a random
            amount between -jitter and jitter
        sample_rate:
            The rate at which waveforms are sampled in Hz
    """

    def __init__(self, jitter: float, sample_rate: float):
        super().__init__()
        self.jitter = jitter
        self.sample_rate = sample_rate
        self.pad = int(self.jitter * self.sample_rate)

    def forward(self, waveforms: torch.Tensor):
        # some array magic to shift waveforms
        # in time domain by different amounts
        shifts = torch.rand(waveforms.size(0), device=waveforms.device)
        shifts = 2 * self.jitter * shifts - self.jitter
        shifts *= self.sample_rate
        shifts = shifts.long()
        indices = (
            torch.arange(waveforms.size(-1), device=waveforms.device)
            .unsqueeze(0)
            .unsqueeze(0)
        )
        indices = indices.expand(*waveforms.shape)
        indices = indices + self.pad
        shifted = indices + shifts.view(-1, 1, 1)

        waveforms = F.pad(
            waveforms, (self.pad, self.pad), mode="constant", value=0
        )
        waveforms = waveforms.gather(2, shifted)
        return waveforms


class TimeAndPhaseShifter(torch.nn.Module):
    """
    A ``torch.nn.Module`` that augments frequency-domain waveforms
    by time and phase shifts. The output is computed as

    .. math::
        h(f) = \exp(2\pi i f t_c) * \exp(2 i \phi_c) * h(f)  # noqa: W605

    Args:
        time_jitter:
            The maximum magnitude of time to shift waveforms in seconds.
    """

    def __init__(self, time_jitter: float = 1.0):
        super().__init__()
        self.time_jitter = time_jitter

    def forward(self, waveforms: torch.Tensor, frequency: torch.Tensor):
        """
        Returns the augmented waveform array. Both arguments assumed to
        be on same device.

        Args:
            waveforms:
                The waveform array, assumed to be shape
                (batch, num_polarization, num_freqs).
            frequency:
                The frequencies for the waveform, assumed to be
                1-D tensor of shape num_freqs
        """
        if waveforms.shape[-1] != frequency.shape[-1]:
            raise RuntimeError(
                "The dimension of h(f) array must match number of frequencies"
            )
        batch_size, num_polarization, _ = waveforms.shape
        time_shifts = self.time_jitter * torch.rand(
            batch_size, device=waveforms.device
        )
        # random time shift up to time_jitter
        phase_shift = torch.exp(
            1j * 2 * torch.pi * torch.outer(time_shifts, frequency)
        )
        # random phase angle up to two pi
        phase_angle = (
            2 * torch.pi * torch.rand(batch_size, device=waveforms.device)
        )
        phase_shift *= torch.exp(
            1j * 2 * torch.outer(phase_angle, torch.ones_like(frequency))
        )
        # repeat phase shifts across polarizations
        phase_shift = phase_shift.unsqueeze(1).repeat(1, num_polarization, 1)
        return waveforms * phase_shift



import torch

from amplfi.train.losses import VICRegLoss


def test_vicreg():
    batch_1 = -5 * torch.randn(1000000, 2)
    batch_2 = 5 * torch.randn(1000000, 2)

    vicreg = VICRegLoss(
        lambda_param=1, mu_param=1, nu_param=1, max_std=1, eps=1e-4
    )

    loss, (inv_loss, var_loss, cov_loss) = vicreg(batch_1, batch_2)

    # each dimension is generated with unit variance so var loss should be 0
    assert torch.isclose(var_loss, torch.tensor(0.0), atol=1e-2)

    # mse 2 * 5**2
    assert torch.isclose(inv_loss, torch.tensor(50.0), atol=1)

    # covariance should be near 0
    assert torch.isclose(cov_loss, torch.tensor(0.0), atol=1e-2)



from .dense import CoherentDenseEmbedding, NChannelDenseEmbedding
from .multimodal import FrequencyPsd, MultiModal, MultiModalPsd
from .resnet import ResNet
from .similarity import Expander, SimilarityEmbedding



import healpy as hp
from typing import Optional
import numpy as np
from astropy.table import Table
from astropy import units as u

from . import distance


def nest2uniq(nside: int, ipix: int):
    return 4 * nside * nside + ipix


def histogram_skymap(
    ra: np.ndarray,
    dec: np.ndarray,
    dist: Optional[np.ndarray] = None,
    nside: int = 32,
    min_samples_per_pix: int = 15,
) -> Table:
    """Given right ascension declination samples
    and optionally distance samples,
    calculate a HEALPix histogram skymap.

    Args:
        ra:
            Samples of right-ascension like parameter between 0 and 2pi.
            Samples outside this range will
            raise an error with the HEALPix library.
        dec:
            Declination samples between -pi/2 and pi/2.
            Samples outside this range will
            raise an error with the HEALPix library.
        dist:
            Distance samples in Mpc. If provided, will calculate distance
            ansatz parameters, `DISTMU`, `DISTSIGMA`, `DISTNORM`
            for each pixel containing more than `min_samples_per_pix`.
            If not provided, will use default values of
            `np.inf`, `1 Mpc`, and `0 / Mpc^2` respectively.
        nside:
            HEALPix nside parameter
        min_samples_per_pix:
            Minimum number of samples per pixel to
            calculate distance ansatz parameters.
            Otherwise, the default values are used.

    Returns:
        astropy.table.Table: HEALPix histogram skymap
    """

    # convert declination to theta between 0 and pi
    theta = np.pi / 2 - dec

    num_samples = len(ra)

    # calculate number of samples in each pixel
    npix = hp.nside2npix(nside)
    ipix = hp.ang2pix(nside, theta, ra, nest=True)
    uniq, counts = np.unique(ipix, return_counts=True)

    # create empty map and then fill in non-zero pix with density
    # estimated by fraction of total samples in each pixel
    map = np.zeros(npix)
    map[np.in1d(range(npix), uniq)] = counts
    density = map / num_samples
    density /= hp.nside2pixarea(nside)
    density /= u.sr

    uniq_ipix = nest2uniq(nside, np.arange(npix))

    # convert to astropy table
    table = Table(
        [uniq_ipix, density],
        names=["UNIQ", "PROBDENSITY"],
        copy=False,
    )

    # defaults for distance parameters
    mu = np.ones(npix) * np.inf
    sigma = np.ones(npix)
    norm = np.zeros(npix)

    if dist is not None:
        # compute distance ansatz for pixels containing
        # greater than a threshold number
        good_ipix = uniq[counts > min_samples_per_pix]
        dist_mu = []
        dist_sigma = []
        dist_norm = []
        for _ipix in good_ipix:
            _distance = dist[ipix == _ipix]
            _, _m, _s = distance.moments_from_samples_impl(_distance)
            _mu, _sigma, _norm = distance.ansatz_impl(_s, _m)
            dist_mu.append(_mu)
            dist_sigma.append(_sigma)
            dist_norm.append(_norm)

        mu[np.in1d(range(npix), good_ipix)] = np.array(dist_mu)
        sigma[np.in1d(range(npix), good_ipix)] = np.array(dist_sigma)
        norm[np.in1d(range(npix), good_ipix)] = np.array(dist_norm)

    mu *= u.Mpc
    sigma *= u.Mpc
    norm /= u.Mpc**2

    # add distance parameters to table
    table.add_columns(
        [mu, sigma, norm], names=["DISTMU", "DISTSIGMA", "DISTNORM"]
    )
    return table


def calculate_searched_area(
    healpix_map: np.ndarray,
    ra_true: float,
    dec_true: float,
    nside: int = 32,
):
    """Calculate the searched area for a given HEALPix map,
    true right ascension and declination. Also calculate estimates
    of 50% and 90% credible regions

    Args:
        healpix_map:
            HEALPix map
        ra_true:
            True right ascension in radians
        dec_true:
            True declination in radians
        nside:
            HEALPix nside parameter
    """
    theta_true = np.pi / 2 - dec_true
    true_ipix = hp.ang2pix(nside, theta_true, ra_true, nest=True)

    # sort pixels in descending order
    # count number of pixels before hitting the pixel with injection
    # in the sorted array
    sorted_idxs = np.argsort(healpix_map)[::-1]
    num_pix_before_injection = 1 + np.argmax(sorted_idxs == true_ipix)
    searched_area = num_pix_before_injection * hp.nside2pixarea(
        nside, degrees=True
    )
    healpix_cumsum = np.cumsum(healpix_map[sorted_idxs])
    fifty = np.argmin(healpix_cumsum < 0.5) * hp.nside2pixarea(
        nside, degrees=True
    )
    ninety = np.argmin(healpix_cumsum < 0.9) * hp.nside2pixarea(
        nside, degrees=True
    )
    return searched_area, fifty, ninety



import pytest
import torch

from amplfi.train.architectures.embeddings import ResNet
from amplfi.train.architectures.flows import NSF


@pytest.fixture(params=[2, 10])
def param_dim(request):
    return request.param


@pytest.fixture(params=[512, 1024])
def strain_dim(request):
    return request.param


@pytest.fixture(params=[10, 12])
def context_dim(request):
    return request.param


@pytest.fixture(params=[1, 2, 3])
def n_ifos(request):
    return request.param


@pytest.fixture(params=[1, 2, 4])
def num_transforms(request):
    return request.param


def test_coupling_flow(
    param_dim, strain_dim, context_dim, n_ifos, num_transforms
):
    data = torch.randn((100, param_dim))
    strain = (torch.randn((100, n_ifos, strain_dim)), None)

    embedding = ResNet(n_ifos, layers=[1, 1], context_dim=context_dim)

    flow = NSF(
        param_dim,
        embedding,
        passes=2,
        transforms=num_transforms,
    )
    log_likelihoods = flow.log_prob(data, context=strain)

    assert log_likelihoods.shape == (len(data),)


def test_autoregressive_flow(
    param_dim, strain_dim, context_dim, n_ifos, num_transforms
):
    data = torch.randn((100, param_dim))
    strain = (torch.randn((100, n_ifos, strain_dim)), None)

    embedding = ResNet(n_ifos, layers=[1, 1], context_dim=context_dim)

    flow = NSF(
        param_dim,
        embedding,
        transforms=num_transforms,
    )
    log_likelihoods = flow.log_prob(data, context=strain)
    assert log_likelihoods.shape == (len(data),)



import torch
from ml4gw.waveforms import SineGaussian

from .generator import WaveformGenerator


class SGGenerator(WaveformGenerator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sine_gaussian = SineGaussian(self.sample_rate, self.duration)

    def slice_waveforms(self, waveforms: torch.Tensor, waveform_size: int):
        # for sine gaussians, place waveform in center of kernel
        center = waveforms.shape[-1] // 2
        half = waveform_size // 2
        start = center - half
        stop = center + half
        return waveforms[..., start:stop]

    def forward(self, **parameters):
        return self.sine_gaussian(**parameters)



from pathlib import Path

import h5py
import torch

from .sampler import WaveformSampler


def x_per_y(x, y):
    return int((x - 1) // y) + 1


class WaveformLoader(WaveformSampler):
    """
    Torch module for loading waveforms from disk,
    performing train/val/test split, and sampling
    them during training.

    TODO: modify this to sample waveforms from disk, taking
    an index sampler object so that DDP training can sample
    different waveforms for each device.

    Args:
        waveform_file:
            Path to the HDF5 file containing the waveforms
        val_frac:
            Fraction of waveforms to use for validation
    """

    def __init__(
        self,
        *args,
        waveform_file: Path,
        val_frac: float,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.val_frac = val_frac
        self.waveform_file = waveform_file

        with h5py.File(waveform_file) as f:
            self.num_waveforms = len(f["signals"])

        self.waveform_file = waveform_file
        (
            self.train_waveforms,
            self.train_parameters,
        ) = self.get_train_waveforms()

    @property
    def num_val_waveforms(self):
        """Total number of validation waveforms across all devices"""
        return int(self.val_frac * self.num_waveforms)

    @property
    def val_waveforms_per_device(self):
        """Number of validation waveforms per device"""
        world_size, _ = self.get_world_size_and_rank()
        return self.num_val_waveforms // world_size

    @property
    def num_train_waveforms(self):
        """Total number of training waveforms"""
        return self.num_waveforms - self.num_val_waveforms

    def load_signals(self, start, stop):
        """
        Load signals and parameters of specified indices from the dataset
        """
        with h5py.File(self.waveform_file) as f:
            signals = torch.Tensor(f["signals"][start:stop])
            parameters = {}
            for parameter in self.inference_params:
                parameters[parameter] = torch.Tensor(f[parameter][start:stop])

        return signals, parameters

    def get_slice_bounds(self, total, world_size, rank) -> tuple[int, int]:
        """
        Determine waveform indices to load for this device
        given our rank and world size
        """
        per_dev = x_per_y(abs(total), world_size)
        start = rank * per_dev
        stop = (rank + 1) * per_dev
        return start, stop

    def get_train_waveforms(
        self,
    ):
        """
        Returns train waveforms for this device
        """
        world_size, rank = self.get_world_size_and_rank()
        start, stop = self.get_slice_bounds(
            self.num_train_waveforms, world_size, rank
        )
        return self.load_signals(start, stop)

    def get_val_waveforms(self):
        """
        Returns validation waveforms for this device
        """
        world_size, rank = self.get_world_size_and_rank()
        start, stop = self.get_slice_bounds(
            self.num_val_waveforms, world_size, rank
        )
        # start counting from the back for val waveforms
        start, stop = -start, -stop or None
        return self.load_signals(start, stop)

    def get_test_waveforms(self, f, world_size, rank):
        """
        Load test waveforms
        """
        return

    def slice_waveforms(self, waveforms: torch.Tensor):
        """
        Slice waveforms to the desired length;
        **NOTE** it is assumed here that waveforms are centered;
        """
        center = waveforms.shape[-1] // 2
        half = self.waveform_length // 2
        start, stop = center - half, center + half
        return waveforms[:, start:stop]

    def sample(self, X):
        """
        Sample method for generating training waveforms
        """
        N = X.shape[0]

        idx = torch.randperm(len(self.train_waveforms))[:N]
        waveforms = self.train_waveforms[idx]
        parameters = {}
        for k, v in self.train_parameters.items():
            parameters[k] = v[idx]

        cross, plus = waveforms
        polarizations = {"cross": cross, "plus": plus}

        return polarizations, parameters



import torch


class Embedding(torch.nn.Module):
    """
    Dummy base class for embedding networks.

    All embeddings should accept `num_ifos`
    as their first argument. They should also
    define a `context_dim` attribute that returns
    the dimensionality of the output of the network,
    which will be used to instantiate the flow transorms.

    This class obvioulsy isn't necessary, but leaving this
    as a reminder that we may wan't to enforce
    the above behavior more explicitly in the future
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.context_dim = None
